# -*- coding: utf-8 -*-
from .. import runners

class Dummy(object):

    def do(self, task, **opts):
        pass

    def add_task(self, task, **opts):
        pass

    def runner(self, ctx, jobs=1, grid_jobs=1):
        return runners.default(ctx, jobs)

from .sge import SGE
from .slurm import Slurm

SGE #pyflakes
Slurm #pyflakes



# -*- coding: utf-8 -*-
from anadama2 import Workflow

workflow = Workflow(remove_options=["input","output"])
workflow.do("ls /usr/bin/ | sort > [t:global_exe.txt]")
workflow.do("ls $HOME/.local/bin/ | sort > [t:local_exe.txt]")
workflow.do("join [d:global_exe.txt] [d:local_exe.txt] > [t:match_exe.txt]")
workflow.go()



# -*- coding: utf-8 -*-
import os
import tempfile
import shutil
import subprocess
import itertools
import sys

from .helpers import sh
from . import Task

try:
    import cPickle as pickle
except ImportError:
    import pickle
    
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class Document(object):
    """A document that is auto generated from a template. 
    
    :param templates: The document template files (or file)
    :type templates: str or list
    
    :param depends: The list of dependencies. 
    :type depends: list of :class:`anadama2.tracked.Base` or strings
    
    :param targets: The target(s). The document(s) to be generated.
    :type targets: :class:`anadama2.tracked.Base` or string
    
    :param vars: A dictionary of variables used by the template.
    :type vars: dict

    """
    
    def create(self, task):
        raise NotImplementedError()

class PweaveDocument(Document):
    """A document that is auto generated from a template using Pweave and Pandoc 
    
    :keyword templates: The document template files (or file)
    :type templates: str or list
    
    :keyword depends: The list of dependencies. 
    :type depends: list of :class:`anadama2.tracked.Base` or strings
    
    :keyword targets: The target(s). The document(s) to be generated.
    :type targets: :class:`anadama2.tracked.Base` or string
    
    :keyword vars: A dictionary of variables used by the template.
    :type vars: dict
    
    :keyword table_of_contents: If set add table of contents to reports
    :type table_of_contents: bool

    """
    
    def __init__(self, templates=None, depends=None, targets=None, vars=None, table_of_contents=None):
        # allow for a single template or multiple templates
        if templates is not None and isinstance(templates,str):
            templates=[templates]
        self.templates=templates
        self.depends=depends
        
        self.table_of_contents=table_of_contents
        
        # if targets is a single item, save as a list
        if targets is not None and isinstance(targets,str):
            targets=[targets]
            
        self.targets=targets
        self.vars=vars
        
        # set the max number of x tick labels to be shown on plots
        self.max_labels = 65
        # set the max labels for legends
        self.max_labels_legend = 30
        
        # set the location of the final figures and data folders
        # only create if the document has a target
        if self.targets:
            self.figures_folder = os.path.join(os.path.dirname(self.targets[0]),"figures")
            if not os.path.isdir(self.figures_folder):
                os.makedirs(self.figures_folder)
                
            self.data_folder = os.path.join(os.path.dirname(self.targets[0]),"data")
            if not os.path.isdir(self.data_folder):
                os.makedirs(self.data_folder)
        else:
            # if a target is not provided use the variables to try to find the folder locations
            try:
                self.vars=self.get_vars()
            except (IndexError, EOFError):
                self.vars=None
                
            if self.vars:
                self.figures_folder = os.path.join(os.path.dirname(self.vars["targets"][0]),"figures")
                self.data_folder = os.path.join(os.path.dirname(self.vars["targets"][0]),"data")
        
        # check for the required dependencies when using a pweave document
        # only check when creating an instance with a template to run
        if templates is not None:
            try:
                import numpy
            except ImportError:
                sys.exit("Please install numpy for document generation")
                
            try:
                # set non-interactive backend to simplify server install
                import matplotlib
                matplotlib.use('Agg')
                import matplotlib.pyplot as pyplot
            except ImportError:
                sys.exit("Please install matplotlib for document generation")
                
            try:
                output=subprocess.check_output(["pypublish","-h"],stderr=subprocess.STDOUT)
            except EnvironmentError:
                sys.exit("Please install pweave for document generation")
                
            try:
                output=subprocess.check_output(["pdflatex","--help"],stderr=subprocess.STDOUT)
            except EnvironmentError:
                sys.exit("Please install latex which includes pdflatex for document generation")
                
        # copy over the file dependencies to the data folder
        if self.depends and self.data_folder and os.path.isdir(self.data_folder):
            depends_files = list(filter(lambda x: not isinstance(x,Task), self.depends))
            for data_file in depends_files:
                if data_file:
                    shutil.copy(data_file,os.path.join(self.data_folder,os.path.basename(data_file)))

    def create(self, task):
        """ Create the documents specified as targets """

        # get the template extension
        template_extension=os.path.splitext(self.templates[0])[-1]
            
        # get the report and figure extensions based on the target
        report_filename, report_extension=os.path.splitext(self.targets[0])

        # create the temp file to run with when ready to create the document
        temp_directory = tempfile.mkdtemp(dir=os.path.dirname(self.targets[0]))
        temp_template_basename = os.path.join(temp_directory,os.path.basename(report_filename))
        # keep the extension of the template for pweave auto reader function
        temp_template = temp_template_basename + template_extension
        
        # merge the templates into the temp file
        with open(temp_template,"w") as handle:
            for file in self.templates:
                for line in open(file):
                    handle.write(line)
                
        handle.close()
        
        # if variables are provided, then create a pickled file to store these
        # for use when the document is created
        if self.vars is not None:
            # save the depends in the vars set
            self.vars["depends"]=self.depends
            
            # save the targets
            self.vars["targets"]=self.targets
            
            # change directories/files to the full paths (so no longer required in input)
            # this is because pweave does not have an output folder option (just cwd so a ch is needed)
            for variable in self.vars:
                try:
                    if os.path.isdir(self.vars[variable]) or os.path.isfile(self.vars[variable]):
                        self.vars[variable] = os.path.abspath(self.vars[variable])
                except (TypeError, ValueError, NameError):
                    # ignore lists and None values
                    continue
            # create a picked file with the temp template name in the same folder
            pickle.dump(self.vars, open(temp_template_basename+".pkl", "wb"))

        # create the document
        # first move to the directory with the temp output files
        # this will cause all pweave output files to be written to this folder
        # by default the tex output file is written to the same folder as the
        # template file and the pdf files are written to the current working directory
        current_working_directory = os.getcwd()
        os.chdir(temp_directory)
        
        # get the intermediate output file based on the initial template name
        intermediate_template = temp_template_basename+"."+"md"
            
        # process the template based on the extension type
        temp_report = temp_template_basename+"."+report_extension
        
        # set the pandoc command based on if a table of contents will be included
        pandoc_command="pandoc {0} -o {1} --variable=linkcolor:Blue "+\
            "--variable=toccolor:Blue --latex-engine=pdflatex --standalone" 
        if self.table_of_contents:
            pandoc_command+=" --toc"
        
        # run pweave then pandoc to generate document
        
        # call pweave to use class with fix
        from pweave import PwebPandocFormatter, Pweb
        from pweave.readers import PwebScriptReader

        class PwebPandocFormatterFixedFigures(PwebPandocFormatter):
            def make_figure_string_size(self, figname, width, label, caption = ""):
                # new function to fix figure width string to work with pandoc format
                # only use width if pandoc installed is >= v1.16.0
                if self.new_pandoc:
                    figstring="![%s](%s){ width=%s }\n" % (caption, figname, width)
                else:
                    figstring="![%s](%s)\n" % (caption, figname) 
                
                if caption == "":
                    figstring += "\\"
        
                figstring += "\n"
                
                return figstring

            def formatfigure(self, chunk):
                fignames = chunk['figure']
                if chunk["caption"]:
                    caption = chunk["caption"]
                else:
                    caption = ""
                figstring = ""
                
                # increase default figure size
                if not chunk["width"]:
                    chunk["width"]="100%"
        
                if chunk['caption'] and len(fignames) > 0:
                    if len(fignames) > 1:
                        print("INFO: Only including the first plot in a chunk when the caption is set")
                        figstring = self.make_figure_string_size(fignames[0], chunk["width"], chunk["name"], caption)
                        return figstring
        
                for fig in fignames:
                    # original line which duplicates figures commented out and replaced
                    #figstring += self.make_figure_string(fignames[0], chunk["width"], chunk["name"])
                    figstring += self.make_figure_string_size(fig, chunk["width"], chunk["name"])
        
                return figstring

        # capture stdout messages
        original_stdout = sys.stdout
        sys.stdout = capture_stdout = StringIO()
        
        doc = Pweb(temp_template)
        doc.setformat(Formatter = PwebPandocFormatterFixedFigures)
        doc.detect_reader()
        doc.weave()
        
        sys.stdout = original_stdout
        
        sh(pandoc_command.format(intermediate_template, temp_report),log_command=True)()          
        
        # change back to original working directory
        os.chdir(current_working_directory)
        
        # rename to the original target name and location specified
        shutil.copy(temp_report,self.targets[0])
        
        # move the temp figures files
        temp_figures_folder=os.path.join(temp_directory,"figures")
        for file in os.listdir(temp_figures_folder):
            shutil.move(os.path.join(temp_figures_folder,file),os.path.join(self.figures_folder,file))
        
        # remove all of the temp files in the temp folder
        shutil.rmtree(temp_directory)
        
        # remove the figures folder if it is empty
        if len(os.listdir(self.figures_folder)) == 0:
            os.rmdir(self.figures_folder)
        
        # remove the data folder if it is empty
        if len(os.listdir(self.data_folder)) == 0:
            os.rmdir(self.data_folder)
        
        
    def get_vars(self):
        """ Try to get the variables from the pickled file """
        
        # the pickled file will be the same name as the template name
        # the current working directory will hold the pickle file
        # the current working directory is a temp folder
        
        # find the pickle file
        pickle_file = list(filter(lambda x: x.endswith(".pkl"),os.listdir(".")))
        
        vars = pickle.load(open(pickle_file[0],"rb"))
        
        return vars
            
    
    def read_table(self, file, invert=None, delimiter="\t", only_data_columns=None, format_data=None):
        """ Read the table from a text file with the first line the column
        names and the first column the row names. 
        
        :param file: The file to read
        :type file: str
            
        :keyword invert: Invert the table rows/columns after reading
        :type invert: bool
        
        :keyword delimiter: The delimiter present in the file
        :type delimiter: str
        
        :keyword only_data_columns: Remove the header and row names
        :type only_data_columns: bool
        
        :keyword format_data: A function to use to format the data
        :type format_data: function
        
        """
        
        def try_format_data(function, data):
            """ Try to format the data, except use zero """
            
            try:
                formatted_data=function(data)
            except ValueError:
                formatted_data=function(0)
                
            return formatted_data
            
        
        # if not set, format data to floats
        if format_data is None:
            format_data=float
        
        data=[]
        row_names=[]
        with open(file) as file_handle:
            column_names = file_handle.readline().rstrip().split(delimiter)[1:]
            for line in file_handle:
                line=line.rstrip().split(delimiter)
                row_names.append(line[0])
                data.append([try_format_data(format_data, i) for i in line[1:]])
                
        # remove extra columns if not requested
        if only_data_columns is not None:
            column_names=[column_names[i] for i in only_data_columns]
            new_data=[]
            for row in data:
                new_data.append([row[i] for i in only_data_columns])
            data=new_data
                
        return column_names, row_names, data
    
    def sorted_data_numerical_or_alphabetical(self, data):
        """ Sort the data numerically or alphabetically depending on data type """

        # sort the data alphabetically
        sorted_data = sorted(data)
        
        try:
            # allow for NA in keys
            na=False
            if "NA" in data:
                data.remove("NA")
                na=True
            sorted_data = sorted(data, key=float)
            if na:
                sorted_data+=["NA"]
        except ValueError:
            pass        
        
        return sorted_data

    def add_threshold(self, threshold, color, label):
         """  Adds horizontal line to plot as threshold
            :param threshold:  float
            :param color:  string
            :param label:  string
         """
         import matplotlib.pyplot as pyplot

         pyplot.axhline(threshold, color=color)
         pyplot.text(0, int(threshold) + 50, label)

         pyplot.show()


    def plot_stacked_barchart_grouped(self, grouped_data, row_labels, column_labels_grouped, title, 
        ylabel=None, legend_title=None, legend_style="normal", legend=True, legend_size=7):
        """ Plot a stacked barchart with data grouped into subplots
        
        :param grouped_data: A dict of lists containing the grouped data
        :type data: dict
        
        :param row_labels: The labels for the data rows
        :type row_labels: list
        
        :param column_labels_grouped: The labels for the columns grouped
        :type column_labels: dict
        
        :param title: The title for the plot
        :type title: str
        
        :keyword ylabel: The y-axis label
        :type ylabel: str
        
        :keyword legend_title: The title for the legend
        :type legend_title: str
        
        :keyword legend_style: The font style for the legend
        :type legend_style: str
        
        :keyword legend: Display legend
        :type legend: bool
        
        :keyword legend_size: The font size for the legend
        :type legend_size: int
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        
        total_groups=len(grouped_data.keys())
        figure, group_axis = pyplot.subplots(1, total_groups, sharey=True, gridspec_kw = {'wspace':0.02})

        # create a set of custom colors to prevent overlap
        # get only a set number of items to recycle colors through subplots
        custom_colors=list(itertools.islice(self._custom_colors(total_colors=len(row_labels)),len(row_labels)))

        # create a subplot for each group
        group_number=0
        
        # sort the groups prior to plotting
        sorted_group_names = self.sorted_data_numerical_or_alphabetical(grouped_data.keys())
            
        # get the total number of columns for all groups
        total_columns_all_groups=len(list(itertools.chain.from_iterable(column_labels_grouped.values())))
        
        for group_name in sorted_group_names:
            data = grouped_data[group_name]
            bar_plots=[]
            # create a plot for each stacked group
            column_labels=column_labels_grouped[group_name]
            plot_indexes=numpy.arange(len(column_labels))
            y_offset=numpy.array([0.0]*len(column_labels))
            for plot_abundance, color in zip(data, custom_colors):
                bar_plots.append(group_axis[group_number].bar(plot_indexes, plot_abundance, 
                    bottom=y_offset, align="center", color=color))
                # add to the y_offset which is the bottom of the stacked plot
                y_offset=y_offset+plot_abundance
            
            # set the current axis to this groups plot
            pyplot.sca(group_axis[group_number])    
            
            # Add the title, labels, and legend
            pyplot.title(group_name, size=12)
            # only set the ylabels for the first subplot
            if ylabel is not None and group_number == 0:
                pyplot.ylabel(ylabel)
            else:
                pyplot.tick_params(axis="y",which="both",left="off",labelleft="off")
                
            # only label the x-axis if all subplots can have labels
            if total_columns_all_groups <= self.max_labels:
                # move the bottom of the figure for larger xaxis labels
                figure.subplots_adjust(bottom=0.3)
                pyplot.xticks(plot_indexes, column_labels, fontsize=7, rotation="vertical")
            else:
                pyplot.tick_params(axis="x",which="both",bottom="off",labelbottom="off")
            pyplot.yticks(fontsize=7)
            
            group_number+=1
            
        # add the legend to the last subplot
        if legend:
            # reduce the size of the plot to fit in the legend
            figure.subplots_adjust(right=0.75)
                
            pyplot.legend(bar_plots,row_labels, loc="center left", bbox_to_anchor=(1,0.5),
                title=legend_title, frameon=False, prop={"size":legend_size, "style":legend_style})
            
        figure.suptitle(title, fontsize=14)
        
        pyplot.show()

    def plot_grouped_barchart(self, data, row_labels, column_labels, title, 
        xlabel=None, ylabel=None, legend_title=None, yaxis_in_millions=None):
        """ Plot a grouped barchart 
        
        :param data: A list of lists containing the data
        :type data: list
        
        :param row_labels: The labels for the data rows
        :type row_labels: list
        
        :param column_labels: The labels for the columns
        :type column_labels: list
        
        :param title: The title for the plot
        :type title: str
        
        :keyword xlabel: The x-axis label
        :type xlabel: str
        
        :keyword ylabel: The y-axis label
        :type ylabel: str
        
        :keyword legend_title: The title for the legend
        :type legend_title: str
        
        :keyword yaxis_in_millions: Show the y-axis in millions
        :type yaxis_in_millions: bool
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        import matplotlib.ticker as ticker
        
        # create a figure subplot to move the legend
        figure = pyplot.figure()
        subplot=pyplot.subplot(111)
        
        # create a set of custom colors to prevent overlap
        custom_colors=self._custom_colors(total_colors=len(row_labels))
        
        # change the yaxis format if set
        axis = pyplot.gca()
        if yaxis_in_millions:
            # get the max value to determine if decimals should be shown on the label 
            max_value=max([max(row)for row in data])/1000000.0
            if max_value <= 0.5:
                yaxis_format = lambda value, position: "{:,.3f}".format(float(value/1000000.0))
            elif max_value <= 1:
                yaxis_format = lambda value, position: "{:,.2f}".format(float(value/1000000.0))
            elif max_value <= 5:
                yaxis_format = lambda value, position: "{:,.1f}".format(float(value/1000000.0))
            else:
                yaxis_format = lambda value, position: "{:,}".format(int(value/1000000))
            axis.get_yaxis().set_major_formatter(ticker.FuncFormatter(yaxis_format))
        
        # set the width of the bars as each total group width is one
        bar_start_point = numpy.arange(len(column_labels))
        gap = 0.1
        bar_width = (1.0 - gap) / len(data)
        
        # create the grouped barplots with gap offsets
        barplots=[]
        for i, data_set in enumerate(data):
            barplots.append(subplot.bar(bar_start_point + i*bar_width, data_set,
                width=bar_width, color=next(custom_colors)))   
        
        # add labels and title
        if xlabel is not None and len(column_labels) <= self.max_labels:
            pyplot.xlabel(xlabel)
        if ylabel is not None:
            pyplot.ylabel(ylabel)
            
        pyplot.title(title)
        
        # place the xticks in the middle of each group
        if len(column_labels) <= self.max_labels:
            # move the bottom of the figure for larger xaxis labels
            # done first before adjusting the width of the figure
            figure.subplots_adjust(bottom=0.3) 
            pyplot.xticks(bar_start_point + 0.5, column_labels, fontsize=7, rotation="vertical")
        else:
            pyplot.tick_params(axis="x",which="both",bottom="off",labelbottom="off")
        pyplot.yticks(fontsize=7)
        
        # set the limits on the x axis so the edge gaps are correct
        pyplot.xlim(0-gap,len(column_labels))
        
        # reduce the size of the plot to fit in the legend
        subplot_position=subplot.get_position()
        subplot.set_position([subplot_position.x0, subplot_position.y0, 
            subplot_position.width *0.80, subplot_position.height])
        
        subplot.legend(barplots,row_labels,loc="center left", bbox_to_anchor=(1,0.5),
            fontsize=7, title=legend_title, frameon=False)
        
        pyplot.show()  
        
    def plot_scatter(self, data, title, row_labels, xlabel=None, ylabel=None, trendline=None):
        """ Plot a scatter plot 

        :param data: A list of lists containing the data
        :type data: list
        
        :param title: The title for the plot
        :type title: str
        
        :param row_labels: The labels for the data rows
        :type row_labels: list
        
        :keyword xlabel: The x-axis label
        :type xlabel: str
        
        :keyword ylabel: The y-axis label
        :type ylabel: str
        
        :keyword trendline: Add a trendline to the plot
        :type trendline: bool
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        
        # create a figure subplot to move the legend
        figure = pyplot.figure()
        subplot = pyplot.subplot(111)

        plots=[]
        for x,y in data:
            # add a scatter plot
            plots.append(subplot.scatter(x,y))
            
            if trendline:
                # compute linear least squares polynomial fit, returns vector of coefficients
                coeff = numpy.polyfit(x,y,1)
                trendline_function = numpy.poly1d(coeff)
                # add trendline to the plot
                pyplot.plot(x,trendline_function(x))
        if ylabel:
            pyplot.ylabel(ylabel)
        if xlabel:
            pyplot.xlabel(xlabel)
            
        # reduce the size of the plot to fit in the legend and the xlabels
        subplot_position=subplot.get_position()
        subplot.set_position([subplot_position.x0, subplot_position.y0, 
            subplot_position.width *0.80, subplot_position.height])
        
        subplot.legend(plots,row_labels,loc="center left", bbox_to_anchor=(1,0.5),
            fontsize=7, frameon=False)
            
        pyplot.title(title)

        pyplot.show()        
        
    def plot_barchart(self, data, labels=None, title=None, xlabel=None, ylabel=None):
        """ Plot a barchart 
        
        :param data: A list of lists containing the data
        :type data: list

        :keyword labels: The labels for the data rows
        :type labels: list

        :keyword title: The title for the plot
        :type title: str
        
        :keyword xlabel: The x-axis label
        :type xlabel: str
        
        :keyword ylabel: The y-axis label
        :type ylabel: str
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        
        figure = pyplot.figure()
        
        # check for a list of lists
        # if a list of lists of single items is found convert to a list
        # if lists of multiple items are found then issue error
        if isinstance(data[0], list):
            max_length=max([len(row) for row in data])
            if max_length == 1:
                data_list=[row[0] for row in data]
                data=data_list
            else:
                raise ValueError("Provide data to the AnADAMA2 document.plot_barchart as a list of floats or ints.")

        positions=numpy.arange(len(data))
        pyplot.bar(positions, data, align="center")
        if labels:
            pyplot.xticks(positions, labels, rotation="vertical")
            # move the bottom of the figure for larger xaxis labels
            figure.subplots_adjust(bottom=0.3)
        
        if ylabel:
            pyplot.ylabel(ylabel)
        if xlabel:
            pyplot.xlabel(xlabel)
            
        pyplot.title(title)

        pyplot.show() 
        
    def _custom_colors(self,total_colors):
        """ Get a set of custom colors for a matplotlib plot """
        
        from matplotlib import cm

        # create a set of custom colors
        
        # get the max amount of colors for a few different color maps
        terrain=[cm.terrain(i/7.0) for i in range(7)]
        # don't use the last dark2 color as this overlaps with the first terrain color
        dark=[cm.Dark2(i/8.0) for i in range(7)]
        jet=[cm.jet(i/7.0) for i in range(7)]
        nipy_spectral=[cm.nipy_spectral(i/10.0) for i in range(10)]
        set3=[cm.Set3(i/12.0) for i in range(12)]
        
        # select the total numer of color maps based on the total number of colors
        if total_colors <= 7:
            sets=[terrain]
        elif total_colors <= 14:
            sets=[terrain,dark]
        elif total_colors <= 24:
            sets=[terrain,dark,nipy_spectral]
        else:
            sets=[terrain,dark,nipy_spectral,set3]
        
        # return a mixed set of colors from each set used
        # repeat colors if we run out
        for color_set in itertools.cycle(zip(*sets)):
            for color in color_set:
                yield color

  

    def plot_stacked_barchart(self, data, row_labels, column_labels, title,
        xlabel=None, ylabel=None, legend_title=None, legend_style="normal", legend_size=7):
        """ Plot a stacked barchart
        
        :param data: A list of lists containing the data
        :type data: list
        
        :param row_labels: The labels for the data rows
        :type row_labels: list
        
        :param column_labels: The labels for the columns
        :type column_labels: list
        
        :param title: The title for the plot
        :type title: str
        
        :keyword xlabel: The x-axis label
        :type xlabel: str
        
        :keyword ylabel: The y-axis label
        :type ylabel: str
        
        :keyword legend_title: The title for the legend
        :type legend_title: str
        
        :keyword legend_style: The font style for the legend
        :type legend_style: str
        
        :keyword legend_size: The font size for the legend
        :type legend_size: int
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        
        figure = pyplot.figure()
        subplot=pyplot.subplot(111)
        bar_plots=[]
        names=[]
        
        # create a set of custom colors to prevent overlap
        custom_colors=self._custom_colors(total_colors=len(row_labels))
        
        # create a plot for each stacked group
        plot_indexes=numpy.arange(len(column_labels))
        y_offset=numpy.array([0.0]*len(column_labels))
        for name, plot_abundance, color in zip(row_labels, data, custom_colors):
            bar_plots.append(subplot.bar(plot_indexes, plot_abundance, 
                bottom=y_offset, align="center", color=color))
            names.append(name)
            # add to the y_offset which is the bottom of the stacked plot
            y_offset=y_offset+plot_abundance
            
        # Add the title, labels, and legend
        if xlabel is not None and len(column_labels) <= self.max_labels:
            subplot.set_xlabel(xlabel)
        if ylabel is not None:
            subplot.set_ylabel(ylabel)
            
        pyplot.title(title)
        
        if len(column_labels) <= self.max_labels:
            # move the bottom of the figure for larger xaxis labels
            # done first before adjusting the width of the figure
            figure.subplots_adjust(bottom=0.3)
            # add labels
            pyplot.xticks(plot_indexes, column_labels, fontsize=7, rotation="vertical")
        else:
            pyplot.tick_params(axis="x",which="both",bottom="off",labelbottom="off")
        
        # reduce the size of the plot to fit in the legend
        subplot_position=subplot.get_position()
        subplot.set_position([subplot_position.x0, subplot_position.y0, 
            subplot_position.width *0.80, subplot_position.height])
            
        pyplot.yticks(fontsize=7)
        subplot.legend(bar_plots,names,loc="center left", bbox_to_anchor=(1,0.5),
            title=legend_title, frameon=False, prop={"size":legend_size, "style":legend_style})
        
        pyplot.show()
        
    def show_table(self, data, row_labels, column_labels, title, format_data_comma=None,
                   location="center", font=None):
        """ Plot the data as a table 
        
        :param data: A list of lists containing the data
        :type data: list
        
        :param row_labels: The labels for the data rows
        :type row_labels: list
        
        :param column_labels: The labels for the columns
        :type column_labels: list
        
        :param title: The title for the plot
        :type title: str
        
        :keyword format_data_comma: Format the data as comma delimited
        :type format_data_comma: bool
        
        :keyword location: The location for the text in the cell
        :type location: str
        
        :keyword font: The size of the font
        :type font: int
        
        """
        
        import numpy
        import matplotlib.pyplot as pyplot
        from matplotlib.table import Table
        
        # if the option is set to format the data, add commas
        if format_data_comma:
            format_data = [map(lambda x: "{:,}".format(int(x)),row) for row in data]
        else:
            format_data = [map(str,row) for row in data]
        data=format_data
        
        # create a figure in one subplot
        figure, axis = pyplot.subplots()
        axis.set_axis_off()
        
        # create a new table instance
        table = Table(axis, bbox=[0,0,1,1])
    
        total_rows=len(row_labels)
        total_columns=len(column_labels)
    
        height = 1.0 / total_rows
    
        # get the width of the columns based on
        # the length of the labels and values
        max_width_chars = [max(map(len,row_labels))]
        for i in range(total_columns):
            current_values=[str(value) for value in [column_labels[i]]+[row[i] for row in data]]
            max_width_chars.append(max(map(len,current_values)))
        
        # compute the widths for each column
        total_chars=sum(max_width_chars)*1.0
        column_widths=[i/total_chars for i in max_width_chars]
    
        # add column labels
        for i, label in enumerate(column_labels):
            table.add_cell(0, i+1, width=column_widths[i+1], height=height, text=label, loc=location)
    
        # add row labels
        for i, label in enumerate(row_labels):
            table.add_cell(i+1, 0, width=column_widths[0], height=height, text=label, loc=location)
    
        # Add data
        for i, row in enumerate(data):
            for j, value in enumerate(row):
                table.add_cell(i+1, j+1, width=column_widths[j+1], height=height, text=value, loc=location)               

        axis.add_table(table)

        # set the font size for the table
        # first must turn off the auto set font size
        table.auto_set_font_size(False)
        font_size=8
        if total_columns > 5:
            font_size=7
            
        # use the font if provided
        if font is not None:
            font_size=font
            
        table.set_fontsize(font_size)
    
        # add the title
        pyplot.title(title)
        
        # plot the table
        pyplot.show() 
        
    def write_table(self, column_labels, row_labels, data, file):
        """ Write a table of data to a file 
        
        :param column_labels: The labels for the columns
        :type column_labels: list
        
        :param row_labels: The labels for the data rows
        :type row_labels: list

        :param data: A list of lists containing the data
        :type data: list        

        :param file: The file to write the table to
        :type file: str
        
        """
        
        # if the folder for the table does not exist, then create
        if not os.path.isdir(os.path.dirname(file)):
            os.makedirs(os.path.dirname(file))
    
        with open(file, "wb") as file_handle:
            file_handle.write("\t".join(column_labels)+"\n")
            for name, row in zip(row_labels, data):
                file_handle.write("\t".join([name]+[str(i) for i in row])+"\n")
        
    def show_hclust2(self,sample_names,feature_names,data,title,log_scale=True,zscore=False,metadata_rows=None,method="correlation"):
        """ Create a hclust2 heatmap with dendrogram and show it in the document
        
        :param sample_names: The names of the samples
        :type sample_names: list
        
        :param feature_names: The names of the features
        :type feature_names: list 
        
        :param data: A list of lists containing the data
        :type data: list         
        
        :param title: The title for the plot
        :type title: str
        
        :keyword log_scale: Show the heatmap with the log scale
        :type log_scale: bool
        
        :keyword zscore: Apply the zscore to the data prior to clustering
        :type zscore: bool

        :keyword metadata_rows: A list of metadata rows
        :type metadata_rows: list

        :keyword method: The distance function for features
        :type method: str
        
        """
        
        from matplotlib._png import read_png
        import matplotlib.pyplot as pyplot
        import numpy
        
        # apply zscore if requested
        if zscore:
            from scipy import stats
            
            data = stats.zscore(numpy.array(data),axis=1)
        
        # write a file of the data
        handle, hclust2_input_file=tempfile.mkstemp(prefix="hclust2_input",dir=os.getcwd())
        heatmap_file=hclust2_input_file+".png"
        if metadata_rows:
            metadata_legend_file = hclust2_input_file+"_legend.png"

        self.write_table(["# "]+sample_names,feature_names,data,hclust2_input_file)
        
        # increase the dpi for small text
        dpi=300
        label_font="8"

        # compute the aspect ratio based on the number of samples and features
        aspect_ratio=len(sample_names)/(len(feature_names)*1.0)
        command=["hclust2.py","-i",hclust2_input_file,"-o",heatmap_file,"--title",title,
            "--title_font",str(int(label_font)*2),"--cell_aspect_ratio",str(aspect_ratio),
            "--flabel_size", label_font, "--slabel_size", label_font,
            "--colorbar_font_size",label_font,"--dpi",str(dpi),"--f_dist_f",method]
        if log_scale:
            command+=["--log_scale"]
        if metadata_rows:
            command+=["--metadata_rows",",".join(str(i) for i in metadata_rows)]
            command+=["--legend_file", metadata_legend_file]
            if len(metadata_rows) > 10:
                command+=["--metadata_height","0.8"]
            elif len(metadata_rows) > 4:
                command+=["--metadata_height","0.4"]
            elif len(metadata_rows) > 1:
                command+=["--metadata_height","0.1"]
            
        # if more than the max samples, do not include sample labels on the heatmap
        if len(sample_names) > self.max_labels:
            command+=["--no_slabels"]
            
        # if more than max labels, do not include the feature labels on the heatmap
        if len(feature_names) > self.max_labels:
            command+=["--no_flabels"]
           
        try: 
            output=subprocess.check_output(command)
            # read the heatmap png file
            heatmap=read_png(heatmap_file)
        except (subprocess.CalledProcessError, OSError):
            print("Unable to generate heatmap")
            heatmap=[]

        # create a subplot and remove the frame and axis labels
        # set the figure and increase the dpi for small text

        fig = pyplot.figure(figsize=(8,8),dpi=dpi)

        if metadata_rows:
            subplot1 = pyplot.subplot2grid((4,1),(0,0), rowspan=3, frame_on=False)
            subplot1.xaxis.set_visible(False)
            subplot1.yaxis.set_visible(False)
        else:
            subplot = fig.add_subplot(111, frame_on=False)
            subplot.xaxis.set_visible(False)
            subplot.yaxis.set_visible(False)
        # show but do not interpolate (as this will make the text hard to read)
        pyplot.imshow(heatmap, interpolation="none")

        if metadata_rows:
            heatmap_legend = read_png(metadata_legend_file)
            # metadata legend subplot
            subplot2 = pyplot.subplot2grid((4,1),(3,0), rowspan=1, frame_on=False)
            subplot2.xaxis.set_visible(False)
            subplot2.yaxis.set_visible(False)
            pyplot.imshow(heatmap_legend, interpolation="none")

        pyplot.show()
        # adjust the heatmap to fit in the figure area
        # this is needed to increase the image size (to fit in the increased figure)
        pyplot.tight_layout()
        
    def _run_r(self, commands, args=None):
        """ Run R on the commands providing the arguments """
        
        if args is None:
            args=[]
        
        proc=subprocess.Popen(["R","--vanilla","--quiet","--args"]+args,
            stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        out, err = proc.communicate(input="\n".join(commands))
        
    def filter_zero_rows(self, row_names, data):
        """ Filter the rows from the data set that sum to zero 
        
        :param row_names: The names of the rows
        :type row_names: list
        
        :param data: A list of lists containing the data
        :type data: list    
        
        """
         
        new_names=[]
        new_data=[]
        for name, row in zip(row_names, data):
            if sum(row) != 0:
                new_names.append(name)
                new_data.append(row)
            
        return new_names, new_data
    
    def filter_zero_columns(self, column_names, data):
        """ Filter the columns from the data set that sum to zero 
        
        :param column_names: The names of the columns
        :type column_names: list
        
        :param data: A list of lists containing the data
        :type data: list    
        
        """

        import numpy
         
        new_names, new_data = self.filter_zero_rows(column_names, numpy.transpose(data))
        data_temp = []
        for row in numpy.transpose(new_data):
            data_temp.append(list(row))
            
        new_data = data_temp
            
        return new_names, new_data
       
    def compute_pcoa(self, sample_names, feature_names, data, apply_transform):
        """ Use the vegan package in R to compute a PCoA. 
        Input data should be organized with samples as columns and features as rows. 
        Data should be scaled to [0-1] if transform is to be applied.
        
        :param sample_names: The labels for the columns
        :type sample_names: list

        :param feature_names: The labels for the data rows
        :type feature_names: list

        :param data: A list of lists containing the data
        :type data: list
        
        :keyword apply_transform: Arcsin transform to be applied
        :type apply_transform: bool
        """

        r_vegan_pcoa=[
            "library(vegan)",
            "args<-commandArgs(TRUE)",
            "data<-read.table(args[1],sep='\\t',header=TRUE, row.names=1)",
            "data.t<-as.data.frame(t(data))"]
        if apply_transform:
            r_vegan_pcoa+=["pcoa<-capscale(asin(sqrt(data.t))~1,distance='bray')"]
        else:
            r_vegan_pcoa+=["pcoa<-capscale(data.t~1,distance='bray')"]
        r_vegan_pcoa+=[
            "write.table(head(eigenvals(pcoa)/sum(eigenvals(pcoa))),args[2],sep='\\t')",
            "write.table(as.data.frame(scores(pcoa,display='sites')),args[3],sep='\\t')"]
        
        # test that the data is scaled to [0-1]
        if apply_transform:
            for row in data:
                out_of_range=list(filter(lambda x: x < 0 or x > 1, row))
                if len(out_of_range) > 0:
                    raise ValueError("Provide data to the AnADAMA2 document.show_pcoa function in the range of [0-1].")
            
        # test for duplicate feature names
        feature_set=set(feature_names)
        if len(list(feature_set)) < len(feature_names):
            raise ValueError("Do not provide duplicate feature names to document.show_pcoa.")
        
        # test samples are provided as the columns of the data
        if len(data[0]) != len(sample_names):
            raise ValueError("Provide data to the AnADAMA2 document.show_pcoa function in the form of samples as columns.")
        
        # test features are provided as rows of the data
        if len(data) != len(feature_names):
            raise ValueError("Provide data to the AnADAMA2 document.show_pcoa function in the form of features as rows.")
        
        # remove any samples from the data for which all features are zero
        sample_names, data = self.filter_zero_columns(sample_names, data)

        # remove any features from the data for which all samples have zero values
        feature_names, data = self.filter_zero_rows(feature_names, data)

        # write a file of the data
        handle, vegan_input_file=tempfile.mkstemp(prefix="vegan_input",dir=os.getcwd())
        eigenvalues_file=vegan_input_file+".eigen"
        scores_file=vegan_input_file+".scores"
        self.write_table(["# "]+sample_names,feature_names,data,vegan_input_file)
        
        self._run_r(r_vegan_pcoa,[vegan_input_file,eigenvalues_file,scores_file])
        
        # get the x and y labels
        r_run_error=False
        try:
            columns, rows, data = self.read_table(eigenvalues_file)
        except EnvironmentError:
            print("No eigenvalues found")
            data=[[0],[0]]
            r_run_error=True
        pcoa1_x_label=int(data[0][0]*100)
        pcoa2_y_label=int(data[1][0]*100)
        
        # get the scores to plot
        try:
            columns, rows, pcoa_data = self.read_table(scores_file)
        except EnvironmentError:
            print("No scores found")
            r_run_error=True
            columns=[]
            rows=[]
            pcoa_data=[]
       
        # if there were no errors, remove the temp files
        if not r_run_error:
            try:
                os.remove(vegan_input_file)
                os.remove(eigenvalues_file)
                os.remove(scores_file)
            except EnvironmentError:
                print("Warning: Unable to remove temp files")

        return pcoa_data, pcoa1_x_label, pcoa2_y_label
            
    def show_pcoa_multiple_plots(self, sample_names, feature_names, data, title, abundances, legend_title="% Abundance", sample_types="samples", feature_types="species", apply_transform=False):
        """ Use the vegan package in R plus matplotlib to plot a PCoA. 
        Input data should be organized with samples as columns and features as rows. 
        Data should be scaled to [0-1] if transform is to be applied.
        Show multiple PCoA plots as subplots each with coloring based on abundance.
        
        :param sample_names: The labels for the columns
        :type sample_names: list

        :param feature_names: The labels for the data rows
        :type feature_names: list

        :param data: A list of lists containing the data
        :type data: list
        
        :param title: The title for the plot
        :type title: str
        
        :param abundances: The sets of abundance data and names for the subplots
        :type abundances: dict
        
        :keyword legend_title: The title for the legend
        :type legend_title: str

        :keyword sample_types: What type of data are the columns
        :type sample_types: str
        
        :keyword feature_types: What type of data are the rows
        :type feature_types: str
        
        :keyword apply_transform: Arcsin transform to be applied
        :type apply_transform: bool
        """

        import numpy
        import matplotlib.pyplot as pyplot
        from matplotlib import cm

        pcoa_data, pcoa1_x_label, pcoa2_y_label=self.compute_pcoa(sample_names, feature_names, data, apply_transform)         
 
        # create a figure and subplots
        nrows = len(abundances.keys())/2
        figure, axis = pyplot.subplots(nrows=nrows,ncols=2)
        # if needed, modify matrix of axis to list
        reformatted_axis = []
        if isinstance(axis[0],numpy.ndarray):
            for axis_list in axis:
                reformatted_axis+=axis_list.tolist()
            axis=reformatted_axis
        figure.suptitle(title,fontsize=12,y=1.002)      
 
        x_values = [x for x,y in pcoa_data]
        y_values = [y for x,y in pcoa_data]
        for subplot, abundance_name in zip(axis,sorted(abundances.keys())):
            pcoa_plot=subplot.scatter(x_values,y_values,c=abundances[abundance_name],cmap=cm.jet)
            figure.colorbar(pcoa_plot,ax=subplot,label=legend_title)
            subplot.set_title(abundance_name)
            subplot.set(xlabel="PCoA 1 ("+str(pcoa1_x_label)+" %)",ylabel="PCoA 2 ("+str(pcoa2_y_label)+" %)")
            subplot.tick_params(axis="both",bottom="off",labelbottom="off",left="off",labelleft="off")
             
        # adjust spacing between subplots
        figure.tight_layout()   

        pyplot.show()
 


    def show_pcoa(self, sample_names, feature_names, data, title, sample_types="samples", feature_types="species",
                  metadata=None, apply_transform=False, sort_function=None, metadata_type=None):
        """ Use the vegan package in R plus matplotlib to plot a PCoA.
        Input data should be organized with samples as columns and features as rows.
        Data should be scaled to [0-1] if transform is to be applied.

        :param sample_names: The labels for the columns
        :type sample_names: list

        :param feature_names: The labels for the data rows
        :type feature_names: list

        :param data: A list of lists containing the data
        :type data: list

        :param title: The title for the plot
        :type title: str

        :keyword sample_types: What type of data are the columns
        :type sample_types: str

        :keyword feature_types: What type of data are the rows
        :type feature_types: str

        :keyword metadata: Metadata for each sample
        :type metadata: dict

        :keyword metadata_type: Type of metadata (continuous or categorical)
        :type metadata_type: str

        :keyword apply_transform: Arcsin transform to be applied
        :type apply_transform: bool

        :keyword sort_function: The function to sort the plot data
        :type sort_function: lambda
        """

        import matplotlib.pyplot as pyplot
        import matplotlib.colors as mcolors
        import matplotlib.cm as cm
        import matplotlib.patches as mpatches
        import numpy as np

        pcoa_data, pcoa1_x_label, pcoa2_y_label = self.compute_pcoa(sample_names, feature_names, data, apply_transform)

        # create a figure subplot to move the legend
        figure = pyplot.figure()
        subplot = pyplot.subplot(111)
        nancolor="grey"

        # create a set of custom colors to prevent overlap
        if metadata:

            metadata_categories = list(set(metadata.values()))
            custom_colors = self._custom_colors(total_colors=len(metadata_categories))

            if metadata_type == 'con':

                cleaned_array = [value for value in metadata_categories if ~np.isnan(value)]
                normalize = mcolors.Normalize(vmin=min(cleaned_array), vmax=max(cleaned_array))
                colormap = pyplot.get_cmap('jet')
                scalarmappaple = cm.ScalarMappable(norm=normalize, cmap=colormap)
                scalarmappaple.set_array(cleaned_array)

                custom_colors_cont = []
                for value in metadata_categories:
                    if np.isnan(value):
                        custom_colors_cont.append(nancolor)
                    else:
                        custom_colors_cont.append(colormap(normalize(value)))

                colors_by_metadata = dict((key, color) for key, color in zip(metadata_categories, custom_colors_cont))

            else:
                colors_by_metadata = dict((key, color) for key, color in zip(metadata_categories, custom_colors))
                colors_by_metadata["NA"] = nancolor

        else:
            custom_colors = self._custom_colors(total_colors=len(pcoa_data))


        # reduce the size of the plot to fit in the legend
        subplot_position = subplot.get_position()
        subplot.set_position([subplot_position.x0, subplot_position.y0,
                              subplot_position.width * 0.80, subplot_position.height])

        plots = []
        metadata_plots = {}
        for i, (x, y) in enumerate(pcoa_data):
            if metadata:
                if metadata[sample_names[i]] not in metadata_plots:
                    metadata_plots[metadata[sample_names[i]]] = [[x], [y]]
                else:
                    metadata_plots[metadata[sample_names[i]]][0].append(x)
                    metadata_plots[metadata[sample_names[i]]][1].append(y)
            else:
                plots.append(subplot.scatter(x, y, color=next(custom_colors)))

        # order the plots alphabetically or numerically
        if not sort_function:
            metadata_ordered_keys = self.sorted_data_numerical_or_alphabetical(metadata_plots.keys())
        else:
            metadata_ordered_keys = sort_function(metadata_plots.keys())

        for key in metadata_ordered_keys:
                plots.append(subplot.scatter(metadata_plots[key][0], metadata_plots[key][1],
                                         color=colors_by_metadata[key]))


        pyplot.title(title)
        pyplot.xlabel("PCoA 1 (" + str(pcoa1_x_label) + " %)")
        pyplot.ylabel("PCoA 2 (" + str(pcoa2_y_label) + " %)")

        # remove the tick marks on both axis
        pyplot.tick_params(axis="x", which="both", bottom="off", labelbottom="off")
        pyplot.tick_params(axis="y", which="both", left="off", labelleft="off")

        if not metadata and len(sample_names) <= self.max_labels_legend:
            subplot.legend(plots, sample_names, loc="center left", bbox_to_anchor=(1, 0.5),
                           fontsize=7, title="Samples", frameon=False)

        if metadata:
            if metadata_type == 'con':
                subplot.append = pyplot.colorbar(scalarmappaple)
                if nancolor in custom_colors_cont:
                    figure.text(0.24, 0.01, "NA/Unknown values are shown in grey.")

            else:
                if len(metadata_ordered_keys) <= self.max_labels_legend:
                    subplot.legend(plots, metadata_ordered_keys, loc="center left", bbox_to_anchor=(1, 0.5),
                               fontsize=7, frameon=False)


        if apply_transform:
            caption = "\n".join(
                ["Principal coordinate analysis of variance among " + sample_types + ", based on Bray-Curtis ",
                 "dissimilarities between " + feature_types + " profiles of " + sample_types + ".  Filtered " + feature_types + "' relative abundances ",
                 "were arcsin-square root transformed to approximate a normal distribution and down-weigh the effect ",
                 "of highly abundant " + feature_types + " on Bray-Curtis dissimilarities.  Numbers in parenthesis on each axis ",
                 "represent the amount of variance explained by that axis."])
        else:
            caption = "\n".join(
                ["Principal coordinate analysis of variance among " + sample_types + ", based on Bray-Curtis ",
                 "dissimilarities between " + feature_types + " profiles of " + sample_types + ".  Numbers in parenthesis on each axis ",
                 "represent the amount of variance explained by that axis."])

        pyplot.show()

        return caption




# -*- coding: utf-8 -*-
import os
import re
import sys
import logging
import itertools
import time

import six

from .grid import Grid
from .grid import GridWorker
from .grid import GridQueue

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

class Slurm(Grid):
    """This class enables the Workflow class to dispatch tasks to
    SLURM. Use it like so:

    .. code:: python

      from anadama2 import Workflow
      from anadama2.slurm import Slurm

      powerup = Slurm(partition="general")
      ctx = Workflow(grid=powerup)
      ctx.do("wget "
             "ftp://public-ftp.hmpdacc.org/"
             "HMMCP/finalData/hmp1.v35.hq.otu.counts.bz2 "
             "-O @{input/hmp1.v35.hq.otu.counts.bz2}")

      # run on slurm with 200 MB of memory, 4 cores, and 60 minutes
      t1 = ctx.grid_do("pbzip2 -d -p 4 < #{input/hmp1.v35.hq.otu.counts.bz2} "
                       "> @{input/hmp1.v35.hq.otu.counts}",
                       mem=200, cores=4, time=60)

      # run on slurm on the serial_requeue partition
      ctx.grid_add_task("some_huge_analysis {depends[0]} {targets[0]}",
                        depends=t1.targets, targets="output.txt",
                        mem=4000, cores=1, time=300, partition="serial_requeue")


      ctx.go()


    :param partition: The name of the SLURM partition to submit tasks to
    :type partition: str

    :param tmpdir: A directory to store temporary files in. All
      machines in the cluster must be able to read the contents of
      this directory; uses :mod:`anadama2.picklerunner` to create
      self-contained scripts to run individual tasks and calls
      ``srun`` to run the script on the cluster.
    :type tmpdir: str
    
    :keyword benchmark_on: Option to turn on/off benchmarking
    : type benchmark_on: bool
    
    :keyword options: Grid specific options to apply to each job
    :type options: str

    """

    def __init__(self, partition, tmpdir, benchmark_on=None, options=None, environment=None):
        super(Slurm, self).__init__("slurm", GridWorker, SLURMQueue(partition, benchmark_on, options, environment), tmpdir, benchmark_on)


class SLURMQueue(GridQueue):
    
    def __init__(self, partition, benchmark_on=None, options=None, environment=None):
        super(SLURMQueue, self).__init__(partition, benchmark_on)
        
        self.options=options
        self.environment=environment
        
        self.job_code_completed="COMPLETED"
        self.job_code_cancelled="CANCELLED"
        self.job_code_failed="FAILED"
        self.job_code_timeout="TIMEOUT"
        self.job_code_memkill="MEMKILL"
       
        self.all_failed_codes=[self.job_code_failed,self.job_code_timeout,self.job_code_memkill,self.job_code_cancelled]
        self.all_stopped_codes=[self.job_code_completed]+self.all_failed_codes
    
    @staticmethod
    def submit_command(grid_script):    
        return ["sbatch",grid_script]
    
    def submit_template(self):
        template = [
            "#SBATCH -p ${partition}",
            "#SBATCH -N 1 ",
            "#SBATCH -n ${cpus}",
            "#SBATCH -t ${time}",
            "#SBATCH --mem ${memory}",
            "#SBATCH -o ${output}",
            "#SBATCH -e ${error}"]
        
        # add user supplied options if provided
        if self.options:
            template+=["#SBATCH "+option for option in self.options]
            
        # add user supplied environment commands if provided
        if self.environment:
            template+=self.environment
            
        return template
    
    def job_failed(self,status):
        # check if the job has a status that it failed
        # This will capture "CANCELLED by 0" and the short form "CANCELLED+"
        return True if status.startswith(self.job_code_cancelled) or status in self.all_failed_codes else False
        
    def job_stopped(self,status):
        # check if the job has a status which indicates it stopped running
        # This will capture "CANCELLED by 0" and the short form "CANCELLED+"
        return True if status.startswith(self.job_code_cancelled) or status in self.all_stopped_codes else False
        
    def job_memkill(self, status, jobid, memory):
        return True if status == self.job_code_memkill else False
        
    def job_timeout(self, status, jobid, time):
        return True if status == self.job_code_timeout else False
    
    def get_job_status_from_stderr(self, error_file, grid_job_status, grid_jobid):
        # read the error file to see if any time or memory errors were reported
        try:
            slurm_errors=subprocess.check_output(["grep","-i","slurmstepd: error\|killed",error_file]).split("\n")
        except (EnvironmentError, subprocess.CalledProcessError):
            slurm_errors=[]
                
        if slurm_errors:
            # check for time or memory
            if list(filter(lambda x: "TIME LIMIT" in x and self.job_code_cancelled in x, slurm_errors)):
                logging.info("Slurm task %s cancelled due to time limit", grid_jobid)
                # This has the slurm status of "TIMEOUT" from slurm sacct
                grid_job_status=self.job_code_timeout
            elif list(filter(lambda x: "exceeded memory limit" in x and "being killed" in x, slurm_errors)) or \
                all([i in "\n".join(slurm_errors).lower() for i in ["exceeded","memory limit","killed"]]):
                logging.info("Slurm task %s cancelled due to memory limit", grid_jobid)
                # This has the slurm status of "CANCELLED by 0" from slurm sacct (short form is "CANCELLED+")
                # It might also have the slurm status of FAILED
                grid_job_status=self.job_code_memkill
                
        return grid_job_status

    def refresh_queue_status(self):
        """ Get the latest status for the grid jobs using the same command for
        jobs in the queue and for completed jobs to benchmark """
        
        # Get the jobid, state, and resources for all jobs for the current user
        stdout=self.run_grid_command_resubmit(["sacct","-o","JobID,State,AllocCPUs,Elapsed,MaxRSS"])

        # remove the header information from the status lines
        # split each line and remove empty lines
        try:
            info=filter(lambda x: x, [line.rstrip().split() for line in stdout.split("\n")[2:]])
        except IndexError:
            info=[]
            
        # now merge the lines for each job so there is only a single item for each job
        # the batch line includes the final MAXRSS
        merged_info={}
        for line in info:
            if not "." in line[0]:
                merged_info[line[0]]=line
            elif ".ba" in line[0]:
                try:
                    merged_info[line[0].split(".")[0]].append(line[-1])
                except KeyError:
                    pass
        info=merged_info.values()

        return list(info)
    



# -*- coding: utf-8 -*-
from anadama2 import Workflow

# create a workflow instance, providing the version number and description
# the version number will appear when running this script with the "--version" option
# the description will appear when running this script with the "--help" option
workflow = Workflow(version="0.1", description="A workflow to run KneadData")

# add the custom arguments to the workflow
workflow.add_argument("kneaddata-db", desc="the kneaddata database", default="/work/code/kneaddata/db/")
workflow.add_argument("input-extension", desc="the input file extension", default="fastq")
workflow.add_argument("threads", desc="number of threads for knead_data to use", default=1)

# get the arguments from the command line
args = workflow.parse_args()

# get all input files with the input extension provided on the command line
in_files = workflow.get_input_files(extension=args.input_extension)

# get a list of output files, one for each input file, with the kneaddata tag
out_files = workflow.name_output_files(name=in_files, tag="kneaddata")

# create a task for each set of input and output files to run kneaddata
workflow.add_task_group(
    "kneaddata --input [depends[0]] --output [output_folder] --reference-db [kneaddata_db] --threads [threads]",
    depends=in_files,
    targets=out_files,
    output_folder=args.output,
    kneaddata_db=args.kneaddata_db,
    threads=args.threads)

workflow.go()



# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import json
import optparse
import itertools

import six

from . import backends
from .util import Bag


desc = "Manipulate the AnADAMA dependency database"

def dump_dependencies(backend):
    for key in backend.keys():
        b = Bag()
        b.name = key
        six.print_(key, "\t", json.dumps(backend.lookup(b)))


def forget(backend, key=None):
    keys = []
    if key:
        keys.append(key)
    if sys.stdin.isatty():
        stdin_keys = iter(line.strip() for line in sys.stdin)
        keys = itertools.chain(keys, stdin_keys)
    for key in keys:
        try:
            backend.delete(key)
        except Exception as e:
            sys.stderr.write("Error inserting key {}: {}\n".format(key, e))


def entry_point(argv=None):
    parser = optparse.OptionParser(description=desc)
    parser.add_option("-d", "--dump-dependencies", action="store_true",
                      help="Print all known dependencies to standard out.")
    parser.add_option("-f", "--forget",
                      help="""Remove dependency from storage backend. Also accepts
                      keys line-by-line from standard in""")
    parser.add_option("-b", "--backend-dir", default=None,
                      help="Specify explicitly where to look for backend data")
    opts, _ = parser.parse_args(args=argv)
    if not opts.backend_dir:
        be_dir = backends.discover_data_directory()
    else:
        be_dir = opts.backend_dir
    backend = backends.auto(be_dir)
    if not backend.exists():
        msg = "Error - backend doesn't exist or is unrecognized: "
        sys.stderr.write(msg+be_dir+"\n")
        sys.exit(1)
    if opts.dump_dependencies:
        return dump_dependencies(backend)
    if opts.forget:
        return forget(backend, opts.forget)



if __name__ == '__main__':
    entry_point(sys.argv)



# -*- coding: utf-8 -*-
from __future__ import print_function
import re
import os
import sys
import json
import zlib
import errno
import inspect
import fnmatch
import mimetypes
import contextlib
import unicodedata
from functools import wraps
from collections import namedtuple
from multiprocessing import cpu_count

import six
from six.moves import zip_longest

from .. import Task

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess


max_cpus = max(1, cpu_count()-1)

class SparseMetadataException(ValueError):
    pass

biopython_to_metaphlan = {
    "fasta": "multifasta",
    "fastq": "multifastq",
    "bam"  : "bam",
}


def first(iterable):
    return next(iter(iterable))


def intatleast1(n):
    return max(1, int(n))


def generator_flatten(gen):
    for item in gen:
        if inspect.isgenerator(item) or type(item) in (list, tuple):
            for value in generator_flatten(item):
                yield value
        else:
            yield item


def guess_seq_filetype(guess_from):
    guess_from = os.path.split(guess_from)[-1]
    if re.search(r'\.f.*q(\.gz|\.bz2)?$', guess_from): #fastq, fnq, fq
        return 'fastq'
    elif re.search(r'\.f.*a(\.gz|\.bz2)?$', guess_from): #fasta, fna, fa
        return 'fasta'
    elif guess_from.endswith('.sff'):
        return 'sff'
    elif guess_from.endswith('.bam'):
        return 'bam'
    elif guess_from.endswith('.sam'):
        return 'sam'


def dict_to_cmd_opts_iter(opts_dict,
                          shortsep=" ",  longsep="=",
                          shortdash="-", longdash="--"):
    """sep separates long options and their values, singlesep separates
    short options and their values e.g. --long=foobar vs -M 2

    """
    if longsep is None:
        longkv = lambda k, v: (longdash+k, v)
    else:
        longkv = lambda k, v: longdash + k + longsep + v

    if shortsep is None:
        shortkv = lambda k, v: (shortdash+k, v)
    else:
        shortkv = lambda k, v: shortdash + k + shortsep + v

    for key, val in opts_dict.items():
        kv = longkv if len(key) > 1 else shortkv
        if val is False or None:
            continue
        elif val is True:
            yield longdash+key if len(key) > 1 else shortdash+key
        elif type(val) in (tuple, list):
            for subval in val:
                yield kv(key, subval)
        else:
            yield kv(key, str(val))

        
def dict_to_cmd_opts(*args, **kwds):
    return " ".join(dict_to_cmd_opts_iter(*args, **kwds))


def mkdirp(path):
    try:
        if not path.startswith("s3://"):
            return os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def is_compressed(fname):
    recognized_compression_types = ("gzip", "bzip2")
    return mimetypes.guess_type(fname)[1] in recognized_compression_types


def filter_compressed(fname_list):
    """Return only files that are known to be compressed and in a
    recognized compression format"""
    return [
        (i,fname) for i,fname in enumerate(fname_list)
        if is_compressed(fname)
    ]


def which_compressed_idxs(fname_mtx):
    for i, raw_tuple in enumerate(fname_mtx):
        for j, fname in enumerate(raw_tuple):
            if is_compressed(fname):
                yield i, j


def take(raw_seq_files, index_list):
    return [
        raw_seq_files[i][j] for i, j in index_list
    ]

        
###
# Serialization things

class SerializationError(TypeError):
    pass


def deserialize_csv(file_handle):
    for i, line in enumerate(file_handle):
        cols = line.split('\t')
        if line.strip() == "":
            continue
        if len(cols) < 2:
            raise AttributeError(
                "Improper formatting in file %s - "
                "only %d column(s) found on line %d" % (
                    file_handle.name, len(cols), i)
            )

        yield ( 
            cols[0], 
            [ col.strip() for col in cols[1:] ] 
            )

        
def deserialize_map_file(file_handle):
    """Returns a list of namedtuples according to the contents of the
    file_handle according to the customs of QIIME
    """
    mangle = lambda field: re.sub(r'\s+', '_', field.strip().replace('#', ''))

    header_line = file_handle.readline()
    if not header_line:
        return
    header = [ mangle(s) for s in header_line.split('\t') ]

    cls = namedtuple('Sample', header, rename=True)

    i = 1
    for row in file_handle:
        i+=1
        if row.startswith('#'):
            continue
        try:
            yield cls._make([ r.strip() for r in row.split('\t') ])
        except TypeError:
            raise SparseMetadataException(
                "Unable to deserialize sample-specific metadata:"+\
                " The file %s has missing values at row %i" %(
                    file_handle.name, i)
            )


def serialize_map_file(namedtuples, output_fname):
    namedtuples_iter = iter(namedtuples)
    with open(output_fname, 'w') as map_file:
        first = next(namedtuples_iter)
        map_file.write(six.u("#"+"\t".join(first._fields)+"\n"))
        map_file.write(six.u("\t".join(first)+"\n"))
        for record in namedtuples_iter:
            map_file.write(six.u("\t".join(record)+"\n"))


def _defaultfunc(obj):
    try:
        return obj._serializable_attrs
    except AttributeError:
        pass

    if hasattr(obj, 'isoformat'):
        return obj.isoformat()
        
    raise SerializationError("Unable to serialize object %s" %(obj))
        

def serialize(obj, to_fp=None):
    if to_fp:
        return json.dump(obj, to_fp, default=_defaultfunc)
    else:
        return json.dumps(obj, default=_defaultfunc)


def deserialize(s=None, from_fp=None):
    if s:
        return json.loads(s)
    elif from_fp:
        return json.load(from_fp)


class SerializableMixin(object):
    """Mixin that defines a few methods to simplify serializing objects
    """

    serializable_attrs = []

    @property
    def _serializable_attrs(self):
        if hasattr(self, "_custom_serialize"):
            return self._custom_serialize()
        else:
            return dict([
                (key, getattr(self, key))
                for key in self.serializable_attrs
            ])


def islambda(func):
    return getattr(func,'func_name') == '<lambda>'


def memoized(func):
    cache = func.cache = {}

    @wraps(func)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = func(*args, **kwargs)
        return cache[args]
    return memoizer


_PATH_list = None
@memoized
def find_on_path(bin_str):
    """ Finds an executable living on the shells PATH variable.
    :param bin_str: String; executable to find

    :returns: Absolute path to `bin_str` or False if not found
    :rtype: str
    """

    global _PATH_list
    if _PATH_list is None:
        _PATH_list = os.environ['PATH'].split(':')

    for dir_ in _PATH_list:
        candidate = os.path.join(dir_, bin_str)
        if os.path.exists(candidate) and os.access(candidate, os.X_OK):
            return candidate

    return False


def partition(it, binsize, pad=None):
    iters = [iter(it)]*binsize
    return zip_longest(fillvalue=pad, *iters)    


def _adler32(fname):
    """Compute the adler32 checksum on a file.

    :param fname: File path to the file to checksum
    :type fname: str
    """
    with open(fname, 'rb') as f:
        checksum = 1
        while True:
            buf = f.read(1024*1024*8)
            if not buf:
                break
            checksum = zlib.adler32(buf, checksum)
            
    return checksum


class ShellException(OSError):
    pass


def sh(cmd, **kwargs):
    kwargs['stdout'] = kwargs.get('stdout', subprocess.PIPE)
    kwargs['stderr'] = kwargs.get('stderr', subprocess.PIPE)
    proc = subprocess.Popen(cmd, **kwargs)
    ret = proc.communicate()
    if proc.returncode:
        msg = "Command `{}' failed. \nOut: {}\nErr: {}"
        raise ShellException(proc.returncode, msg.format(cmd, ret[0], ret[1]))
    return ret


class Bag(object):
    pass


class HasNoEqual(object):
    def __eq__(self, other):
        return False


class Directory(object):
    def __init__(self, name):
        self.name = os.path.abspath(name)

    def files(self, pattern=None):
        names = [ f for f in os.listdir(self.name)
                  if f not in ('.', '..') ]
        if pattern:
            names = fnmatch.filter(names, pattern)
        return names

    def exists(self):
        return os.path.isdir(self.name)

    def create(self):
        try:
            os.mkdir(self.name)
        except OSError as e:
            if e.errno != 17: #already exists
                raise
            
    def __str__(self):
        return self.name+"/"

    def __repr__(self):
        return "Directory('{}/')".format(self.name)


def noop(*args, **kwargs):
    return None


def istask(x):
    return isinstance(x, Task)


def isnottask(x):
    return not isinstance(x, Task)


def underscore(s, repl=r'[\s/:@\,*?]+'):
    """Remove all whitespace and replace with underscores"""
    return re.sub(repl, '_', s)


def sugar_list(x):
    """Turns just a single thing into a list containing that single thing.

    """

    if istask(x) \
       or not hasattr(x, "__iter__") \
       or isinstance(x, six.string_types):
        return [x]
    else:
        return x


def keepkeys(d, keys):
    """Drop all keys from dictionary ``d`` except those in
    ``keys``. Modifies the dictionary in place!
    
    :param d: The dictionary to modify
    :type d: dict

    :param keys: The keys to keep
    :type keys: iterable
    """

    ks = set(list(keys))
    to_rm = [k for k in d.keys() if k not in ks]
    for k in to_rm:
        del d[k]
    return d


def keyrename(d, mapping):
    """Change all keys in dictionary ``d`` according to
    ``mapping``. Modifies ``d`` in place!

    :param d: The dictionary to change
    :type d: dict

    :param mapping: The keys to change and to what to change them
    :type mapping: iterable of 2-tuples

    """

    for frm, to in mapping:
        d[to] = d.pop(frm)
    return d


def dichotomize(it, tf_func):
    """Split an iterable into two lists by use of a function that returns
    True or False.

    :param it: The items to split

    :param tf_func: A function that returns True or False

    :returns: 2-Tuple of lists: the items for which the function
      returned True, and the items for which the function returned
      False

    """
    t, f, = list(), list()
    for item in it:
        if tf_func(item):
            t.append(item)
        else:
            f.append(item)
    return t, f


def kebab(s):
    """Kebab-case a string. Intra-string whitespace is converted to ``-``
    and unfriendly characters are dropped."""
    if type(s) is six.text_type:
        s = unicodedata.normalize("NFKD", s).encode("ascii", "ignore")
        if six.PY3:
            s = s.decode("ascii")
    s = re.sub(r"[\s-]+", '-', s)
    return re.sub(r"""['".,\[\]{}!@#$%^&*()_=+|\\`~><\d]+""", '', s).rstrip('-')


def get_name(t):
    """Get the name of a tracked object

    :param t: The tracked object
    :type t: objects implementing the :class:`anadama2.tracked.Base`
      interface

    """
    return t.name


@contextlib.contextmanager
def capture(stderr=None, stdout=None):
    if stderr:
        saved_stderr = sys.stderr
        sys.stderr = stderr
    if stdout:
        saved_stdout = sys.stdout
        sys.stdout = stdout
    yield
    if stderr:
        sys.stderr = saved_stderr
    if stdout:
        sys.stdout = saved_stdout
    



# -*- coding: utf-8 -*-
"""Contains functions that help create tasks. All functions contained
herein are intended for use with
:meth:`anadama2.workflow.Workflow.add_task`. This means that the functions
in here don't immediately do what they say; they return functions
that, when called, do that they say (they're closures). Sorry if that
breaks your brain.

Using closures lets you add tasks like this:

.. code:: python

  from anadama2 import Workflow
  from anadama2.helpers import sh

  ctx = Workflow()
  ctx.add_task(sh("my fancy shell command"),
               targets="foobaz.txt")

Instead of this:

.. code:: python


  from anadama2 import Workflow
  from anadama2.util import sh # <--- note the different import

  ctx = Workflow()
  ctx.add_task(lambda task: sh("my fancy shell command"),
               targets="foobaz.txt")


"""

import os
import shutil
import logging
import re

import six

from .util import sh as _sh
from .util import sugar_list

from .tracked import try_get_local_path
from .reporters import SHELL_COMMAND

def file_size(depends):
    """ Return the size of the file in GB """
    
    # allow for paths to files and also dependency classes
    file_name = depends
    if hasattr(depends,"name"):
        file_name = depends.name

    try:
        size = os.path.getsize(file_name) / (1024.0**3)
    except (OSError, AttributeError):
        size = 0
        
    return size 


def apply_sh(actions):
    """Add the shell function to any actions that are strings"""
    return [ a if six.callable(a) else sh(a, log_command=False) for a in actions ]


def sh(s, log_command=True, **kwargs):
    """Execute a shell command. All further keywords are passed to
    :class:`subprocess.Popen`

    :param s: The command to execute. Passed directly to a shell, so
      be careful about doing things like ``sh('df -h > data; rm -rf
      /')``; both commands are executed and bad things will happen.
    :type s: str

    """
    def actually_sh(task=None):
        logger = logging.getLogger(__name__)
        if log_command:
            logger.info(SHELL_COMMAND+s)
        kwargs['shell'] = True
        ret = _sh(s, **kwargs)
        logger.info("Execution complete. Stdout: %s\nStderr: %s",
                    ret[0] or '',
                    ret[1] or '')
    return actually_sh

def build_actions(actions, deps, targs, visible, kwds, use_parse_sh=True):
    actions = filter(None, sugar_list(actions))
    if use_parse_sh:
        return [ a if six.callable(a) else format_command(a, depends=deps, targets=targs, **kwds)
                 for a in actions ]
    else:
        return [a for a in actions]


def format_command(command, **kwargs):
    """Format the shell command to allow for special variables 
    Here's a synopsis of
    common use cases:

    - ``[targets[0]]`` is formatted to the first target
    
    - ``[depends[2]]`` is formatted to the third dependency

    Extra keyword arguments are also added to the formatting keyword
    arguments. Thus, adding a keyword argument of ``threads=1`` makes
    ``[threads]`` be formatted to ``1`` in the shell command.
    :type s: str
    """

    # start with the longest keys for replacement first
    keys=sorted(kwargs.keys(), key=len)
        
    # store the original command, incase needed for error message
    original_command = command
        
    # replace instances of "[key]" with value (local path if available)
    # for values that are lists, replace "[key[0]]" with value
    for key in keys:
        replacement=kwargs[key]
        if isinstance(replacement, list) or isinstance(replacement, tuple):
            replacement=map(try_get_local_path,replacement)
            for i, item in enumerate(replacement):
                command=command.replace("["+str(key)+"["+str(i)+"]]",str(item))
            # for lists/tuples with one item, allow for index to not be included
            if len(replacement) == 1:
                command=command.replace("["+str(key)+"]",str(replacement[0]))
        else:
            replacement=try_get_local_path(replacement) 
            command=command.replace("["+str(key)+"]",str(replacement))
            # also allow for the item to be referenced as the first object
            command=command.replace("["+str(key)+"[0]]",str(replacement))
            
    # check for any keywords in the command that were not replaced
    # allow for bash test constructs
    if re.search("\[[a-zA-Z]",command):
        message="Unable to replace all keys in command.  "
        message+="Original command: "+original_command+"  "
        message+="Final formatted command: "+command+"  "
        raise KeyError(message)
    
    return command 

def parse_sh(s, **kwargs):
    """Do the same thing as :func:`anadama2.helpers.sh`, but do some extra
    interpreting and formatting of the shell command before handing it
    over to the shell. For formatting information, see 
    :func:`anadama2.helpers.format_command`.
    :type s: str
    """

    def actually_sh(task):
        fmtd = format_command(s, depends=task.depends, targets=task.targets, **kwargs)
        logger = logging.getLogger(__name__)
        logger.info("Executing with shell: "+fmtd)
        ret = _sh(fmtd, shell=True)
        logger.info("Execution complete. Stdout: %s\nStderr: %s",
                    ret[0] or '',
                    ret[1] or '')
    return actually_sh


def system(args_list, stdin=None, stdout=None, stdout_clobber=None,
           stderr=None, stderr_clobber=None, working_dir=None, 
           **kwargs):
    """Execute a system call (no shell will be used). All further keywords
    are passed to :class:`subprocess.Popen`

    :param args_list: The argv to be passed to the system call.
    :type args_list: list

    :keyword stdin: If provided, the name of the file to open and send
      to the subprocess' standard input. By default no data is sent to
      the process.
    :type stdin: str

    :keyword stdout: If provided, the name of the file to send output
      from the subprocess' standard output. Standard output is
      appended to the file. By default all data from the subprocess
      standard out is sent to the standard out of the executing
      process
    :type stdout: str

    :keyword stdout_clobber: If provided, the name of the file to send
      output from the subprocess' standard output. If the file already
      exists, it will be truncated before it receives writes.
    :type stdout_clobber: str

    :keyword stderr: If provided, the name of the file to send output
      from the subprocess' standard error output. Standard error
      output is appended to the file. By default all data from the
      subprocess standard error is sent to the standard error of the
      executing process.
    :type stderr: str

    :keyword stderr_clobber: If provided, the name of the file to send
      output from the subprocess' standard error output. If the file
      already exists, it will be truncated before it receives writes.
    :type stderr_clobber: str

    """
    kwargs.pop("shell", None)
    args_list = list(map(str, args_list))
    __sh = _sh
    if working_dir is not None:
        def __sh(*a, **kw):
            prev = os.getcwd()
            os.chdir(working_dir)
            try:
                ret = _sh(*a, **kw)
            finally:
                os.chdir(prev)
            return ret
    def actually_system(task):
        files = []
        if stdin:
            f = kwargs['stdin'] = open(stdin, 'rb')
            files.append(f)
        if stdout:
            f = kwargs['stdout'] = open(stdout, 'ab')
            files.append(f)
        if stdout_clobber:
            f = kwargs['stdout'] = open(stdout_clobber, 'wb')
            files.append(f)
        if stderr:
            f = kwargs['stderr'] = open(stderr, 'ab')
            files.append(f)
        if stderr_clobber:
            f = kwargs['stderr'] = open(stderr_clobber, 'wb')
            files.append(f)
        try:
            logger = logging.getLogger(__name__)
            logger.info(
                "Forking subprocess %s with args %s", args_list, kwargs)
            ret = _sh(args_list, **kwargs)
            logger.info("Execution complete. Stdout: %s\nStderr: %s",
                        ret[0] or '',
                        ret[1] or '')
        finally:
            for f in files:
                f.close()
        return ret
    return actually_system


def rm(to_rm, ignore_missing=True):
    """Remove files using :func:`os.remove`.
    
    :param to_rm: The filename or filenames to remove.
    :type to_rm: str or list of str

    :keyword ignore_missing: If one of the filenames isn't a file,
      don't raise an exception
    :type ignore_missing: bool

    """

    def actually_rm(task):
        for f in sugar_list(to_rm):
            if os.path.isfile(f) or not ignore_missing:
                logging.getLogger(__name__).info("Removing "+f)
                os.remove(f)
    return actually_rm


def rm_r(to_rm, ignore_missing=True):
    """Recursively remove files and directories using
    :func:`shutil.rmtree`.
    
    :param to_rm: The filename or filenames to remove.
    :type to_rm: str or list of str

    :keyword ignore_missing: If one of the filenames isn't a file,
      don't raise an exception
    :type ignore_missing: bool

    """
    def actually_rm_r(task):
        for f in sugar_list(to_rm):
            logging.getLogger(__name__).info("Removing recursively: "+f)
            shutil.rmtree(f, ignore_errors=ignore_missing)
    return actually_rm_r




# -*- coding: utf-8 -*-
"""Save a task to a script for running by other programs"""

import os
import sys
import tempfile
import copy

import cloudpickle

template = \
r"""

import cloudpickle

from anadama2.runners import _run_task_locally


# set the variables
in_file = "{in_file}"
out_file = "{out_file}"

# load the task
task = cloudpickle.load(open(in_file,"rb"))

# run the task
result = _run_task_locally(task)

# write the pickled results to a file
cloudpickle.dump(result,open(out_file,"wb"))

"""

class PickleScript(object):
    def __init__(self, task, tmpdir, suffix):
        # create temp files for the script, input and the output file
        output_handle, output_file = tempfile.mkstemp(dir=tmpdir, suffix=suffix+"_output.pkl")
        os.close(output_handle)
        input_handle, input_file = tempfile.mkstemp(dir=tmpdir, suffix=suffix+"_input.pkl")
        os.close(input_handle)
        script_handle, script_file = tempfile.mkstemp(dir=tmpdir, suffix=suffix+"_picklerunner.py")
        os.close(script_handle)
        
        self.task = task
        self.output_file = output_file
        self.input_file = input_file
        self.script_file = script_file
        
    def create_task(self):
        # write the input pickle file
        with open(self.input_file,"wb") as file_handle:
            cloudpickle.dump(self.task,file_handle)
        
        # write the script
        custom_script = template.format(in_file=self.input_file, out_file=self.output_file)
        with open(self.script_file,"wb") as file_handle:
                file_handle.write(custom_script.encode("utf-8"))
        
        # update the task to run the pickle script
        pickle_task = copy.deepcopy(self.task)
        pickle_task.actions = [self.run_command()]
        
        return pickle_task
        
    def run_command(self):
        return sys.executable+" "+self.script_file

    def result(self, result):
        # try to get the result from running the pickled function
        extra_error = None
        try:
            with open(self.output_file,"rb") as file_handle:
                result = cloudpickle.load(file_handle)
        except (ValueError, EOFError):
            extra_error = "Unable to decode pickle task result"
        
        if extra_error:    
            result = result._replace(error=str(result.error)+extra_error)
            
        return result




# -*- coding: utf-8 -*-
import os
import sys
import time
import tempfile
import pwd

import six

from .grid import Grid
from .grid import GridWorker
from .grid import GridQueue

from .. import runners
from .. import picklerunner
from ..util import underscore
from ..util import find_on_path
from ..util import keyrename, keepkeys

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

class SGE(Grid):
    """This class enables the Workflow class to dispatch tasks to
    Sun Grid Engine and its lookalikes. Use it like so:

    .. code:: python

      from anadama2 import Workflow
      from anadama2.sge import SGE

      ctx = Workflow(grid_powerup=SGE(queue="general"))
      ctx.do("wget "
             "ftp://public-ftp.hmpdacc.org/"
             "HMMCP/finalData/hmp1.v35.hq.otu.counts.bz2 "
             "-O @{input/hmp1.v35.hq.otu.counts.bz2}")

      # run on sge with 200 MB of memory, 4 cores, and 60 minutes
      t1 = ctx.grid_do("pbzip2 -d -p 4 < #{input/hmp1.v35.hq.otu.counts.bz2} "
                       "> @{input/hmp1.v35.hq.otu.counts}",
                       mem=200, cores=4, time=60)

      # run on sge on the serial_requeue queue
      ctx.grid_add_task("some_huge_analysis {depends[0]} {targets[0]}",
                        depends=t1.targets, targets="output.txt",
                        mem=4000, cores=1, time=300, partition="serial_requeue")


      ctx.go()


    :param partition: The name of the SLURM partition to submit tasks to
    :type partition: str

    :param tmpdir: A directory to store temporary files in. All
      machines in the cluster must be able to read the contents of
      this directory; uses :mod:`anadama2.picklerunner` to create
      self-contained scripts to run individual tasks and calls
      ``srun`` to run the script on the cluster.
    :type tmpdir: str
    
    :keyword benchmark_on: Option to turn on/off benchmarking
    : type benchmark_on: bool
    
    :keyword options: Grid specific options to apply to each job
    :type options: str

    """

    def __init__(self, partition, tmpdir, benchmark_on=None, options=None, environment=None):
        super(SGE, self).__init__("sge", GridWorker, SGEQueue(partition, benchmark_on, options, environment), tmpdir, benchmark_on)
        

class SGEQueue(GridQueue):
    
    def __init__(self, partition, benchmark_on=None, options=None, environment=None):
        super(SGEQueue, self).__init__(partition, benchmark_on) 
        
        self.options=options
        self.environment=environment
        
        self.job_code_completed="COMPLETED"
        self.job_code_error="FAILED"
        self.job_code_terminated="TERMINATED"
        self.job_code_deleted="DELETED"
       
        self.all_failed_codes=[self.job_code_error,self.job_code_terminated,self.job_code_deleted]
        self.all_stopped_codes=[self.job_code_completed]+self.all_failed_codes
       
        # allow for jobs to be terminated when about to reach the memory requested
        self.memory_buffer = 1024
    
    @staticmethod
    def submit_command(grid_script):    
        return ["qsub",grid_script]
    
    def submit_template(self):
        template = [
            "#$$ -q ${partition}",
            "#$$ -pe smp ${cpus}",
            "#$$ -l h_rt=${time}",
            "#$$ -l h_vmem=${memory}m",
            "#$$ -o ${output}",
            "#$$ -e ${error}"]

        # add user supplied options if provided
        if self.options:
            template+=["#$$ "+option for option in self.options]        
            
        # add user supplied environment commands if provided
        if self.environment:
            template+=self.environment

        return template
    
    def job_failed(self,status):
        # check if the job has a status that it failed
        return True if status in self.all_failed_codes else False
        
    def job_stopped(self,status):
        # check if the job has a status which indicates it stopped running
        return True if status in self.all_stopped_codes else False
    
    def job_memkill(self, status, jobid, memory):
        # check if the job was killed because it used too much memory
        new_status, cpus, new_time, new_memory = self.get_benchmark(jobid)
        
        # if memory is not yet available for the job, wait for a new benchmark
        if new_memory == "NA":
            new_status, cpus, new_time, new_memory = self.get_benchmark(jobid, wait=True)
        
        try:
            exceed_allocation = True if (float(new_memory) + self.memory_buffer) > float(memory) else False
        except ValueError:
            exceed_allocation = False
            
        return True if exceed_allocation and new_status == self.job_code_terminated else False
            
    def job_timeout(self, status, jobid, time):
        # check if the job was killed because it used too much memory
        new_status, cpus, new_time, new_memory = self.get_benchmark(jobid)
        
        # if time is not yet available for the job, wait for a new benchmark
        if new_time == "NA":
            new_status, cpus, new_time, new_memory = self.get_benchmark(jobid, wait=True)
        
        try:
            exceed_allocation = True if float(new_time) > float(time) else False
        except ValueError:
            exceed_allocation = False
            
        return True if exceed_allocation and new_status == self.job_code_terminated else False
            
    def refresh_queue_status(self):
        """ Get the latest status for the grid jobs using the same command for
        jobs in the queue and for completed jobs to benchmark """
        
        # Get the jobid and state for all jobs pending/running/completed for the current user
        qacct_stdout=self.run_grid_command_resubmit(["qacct","-o",pwd.getpwuid(os.getuid())[0],"-j","*"])
        
        # info list should include jobid, state, cpus, time, and maxrss
        info=[]
        job_status=[]
        for line in qacct_stdout.split("\n"):
            if line.startswith("jobnumber") or line.startswith("job_number"):
                if job_status:
                    info.append(job_status)
                job_status=[line.rstrip().split()[-1],"NA","NA","NA","NA"]
            # get the states for completed jobs
            elif line.startswith("failed"):
                failed_code = line.rstrip().split()[1]
                if failed_code != "0":
                    if failed_code in ["37","100"]:
                        job_status[1]=self.job_code_terminated
                    else:
                        job_status[1]=self.job_code_error
            elif line.startswith("deleted_by"):
                if line.rstrip().split()[-1] != "NONE" and job_status[1] == self.job_code_terminated:
                    job_status[1]=self.job_code_deleted
            elif line.startswith("exit_status"):
                # only record if status has not yet been set
                if job_status[1] == "NA":
                    exit_status = line.rstrip().split()[-1]
                    if exit_status == "0":
                        job_status[1]=self.job_code_completed
                    elif exit_status == "137":
                        job_status[1]=self.job_code_terminated
                    else:
                        job_status[1]=self.job_code_error
            # get the current state for running jobs
            elif line.startswith("job_state"):
                job_status[1]=line.rstrip().split()[-1]
            elif line.startswith("slots"):
                job_status[2]=line.rstrip().split()[-1]
            elif line.startswith("ru_wallclock"):
                try:
                    # get the elapsed time in minutes
                    job_status[3]=str(float(line.rstrip().split()[-1])/60.0)
                except ValueError:
                    job_status[3]="NA"
            elif line.startswith("ru_maxrss"):
                job_status[4]=line.rstrip().split()[-1]+"K"
        
        if job_status:
            info.append(job_status)

        return info








# -*- coding: utf-8 -*-
"""utilities for working with filenames"""

import os
import re
import sys
from . import mkdirp

_owd = os.getcwd()
original_wd = lambda: _owd
cwd = os.getcwd

def mangle(fname, tag=None, dir=None, ext=None):
    """Create a new filename from an old filename.

    :param fname: The old filename used as a template for the new
      filename
    
    :keyword tag: Add this tag to the new filename. See
      :func:`anadama2.util.fname.addtag`.
    
    :keyword dir: Change the path of the filename to ``dir``

    :keyword ext: Swap out the extension for ``ext``.

    :returns: The new filename
    """

    new = fname[:]
    if tag: # don't want empty strings
        new = addtag(new, tag)
    if dir is not None:
        new = os.path.join(dir, os.path.basename(new))
    if ext is not None:
        new = addext(rmext(new, all=True), ext)
    return new_file(new)


def _new_file(*names, **opts):
    basedir = opts.get("basedir")
    for name in names:
        if basedir:
            name = os.path.abspath(
                os.path.join(
                    basedir, os.path.basename(name)
                )
            )
        
        dir = os.path.dirname(name)
        if dir and not os.path.exists(dir):
            mkdirp(dir)

        yield name


def new_file(*names, **opts):
    iterator = _new_file(*names, basedir=opts.get("basedir"))
    if len(names) == 1:
        return next(iterator)
    else:
        return list(iterator)


def addtag(name_str, tag_str):

    """Add a tag to a filename. The tag is placed at the end of the
    filename just before the extension. If the filename has no extension,
    it's placed right on the end.

    Here's an example::

      >>> from anadama2.util.fname import addtag
      >>> addtag("myfile.txt", "big")
      'myfile_big.txt'
      >>>

    """
    path, name_str = os.path.split(name_str)
    match = re.match(r'(.+?)(\..*)', name_str)
    if match:
        base, ext = match.groups()
        return os.path.join(path, base + "_" + tag_str + ext)
    else:
        return os.path.join(path, name_str + "_" + tag_str)


def addext(name_str, ext_str):
    """Add a file extension to a filename
    
    :param name_str: The filename that will get the extension

    :param ext_str: The extension (no leading ``.`` required)

    :returns: The filename with the extension
    """

    if not ext_str:
        return name_str
    return name_str + "." + ext_str


def rmext(name_str, all=False):
    """Remove file extensions 

    :param name_str: The filename to remove extensions from
    :type name_str: str or unicode

    :keyword all: remove all extensions.
    :type all: bool

    """

    _match = lambda name_str: re.match(r'(.+)(\..*)', name_str)
    path, name_str = os.path.split(name_str)
    match = _match(name_str)
    while match:
        name_str = match.group(1)
        match = _match(name_str)
        if not all:
            break

    return os.path.join(path, name_str)


def script_wd():
    """Return the directory of script being executed (sys.argv[0]). Most
    of the time, that's the directory that contains the script that
    calls ctx.go()

    """

    return os.path.abspath(os.path.dirname(sys.argv[0]))




# -*- coding: utf-8 -*-
import os
import sys
import logging
import itertools
from glob import glob
from operator import eq, itemgetter
from collections import defaultdict
import subprocess

import six
from six.moves import zip_longest

from .util import _adler32, find_on_path, sh, HasNoEqual
from .util import istask, Directory

logger = logging.getLogger(__name__)
_singleton_idx = defaultdict(dict)
first = itemgetter(0)

def auto(x):
    """Translate a string, function or task into the appropriate subclass
    of :class:`anadama2.tracked.Base`. Tildes and shell
    variables are expanded using :func:`os.path.expanduser` and
    :func:`os.path.expandvars`. If that's not your game, use
    :class:`anadama2.tracked.TrackedDirectory` or
    :class:`anadama2.tracked.HugeTrackedFile`
    as appropriate. The current mapping is as follows:

    - Subclasses of :class:`anadama2.tracked.Base` are returned as is

    - Strings ending in '/' ``->`` :class:`anadama2.tracked.TrackedDirectory`

    - Strings not ending in '/' ``->`` :class:`anadama2.tracked.HugeTrackedFile`

    - Instances of subclasses of :class:`anadama2.Task` are handled
      specially by :meth:`anadama2.workflow.Workflow.add_task` and
      are returned as is

    - Functions or other callables ``->`` :class:`anadama2.tracked.TrackedFunction`

    :param x: The object to be translated into a dependency object

    """

    if isinstance(x, six.string_types):
        return _autostring(x)
    elif istask(x):
        return x
    elif isinstance(x, Base):
        return x
    elif six.callable(x):
        # no explicit key is given, so I have to make one up
        key = "Function ({})".format(x.__name__)
        return TrackedFunction(key, x)
    elif isinstance(x, Directory):
        return TrackedDirectory(x.name)
    else:
        raise ValueError(
            "Not sure how to make `{}' into a dependency".format(x))


def _autostring(s):
    if s.startswith("s3:/"):
        return AWSHugeTrackedFile(s)
    s = os.path.expanduser(os.path.expandvars(s))
    if s.endswith('/'):
        return TrackedDirectory(s)
    else:
        return HugeTrackedFile(s)


def any_different(ds, backend):
    """Determine whether any dependencies have changed since last save.

    :param ds: The dependencies in question
    :type ds: instances of any :class:`anadama2.tracked.Base` subclass

    :param backend: Backend to query past results of a dependency object
    :type backend: instances of any :class:`anadama2.backends.BaseBackend` subclass
    """
    isdebug = logger.isEnabledFor(logging.DEBUG)
    
    for dep in ds:
        past_dep_compare = backend.lookup(dep)
        if not past_dep_compare:
            if isdebug:
                logger.debug("Dep `%s' of type %s changed: "
                             "wasn't previously saved in backend",
                             dep.name, type(dep))
            return True

        compares = zip_longest(
            dep.compare(), past_dep_compare, fillvalue=HasNoEqual())
        the_same = itertools.starmap(eq, compares)
        try:
            is_different = not all(the_same)
        except:
            if isdebug:
                logger.debug("Dep `%s' of type %s changed: "
                             "hit an exception when running .compare()",
                             dep.name, type(dep))
            return True
        if is_different:
            if isdebug:
                logger.debug("Dep `%s' of type %s changed: "
                             ".compare() different since last save",
                             dep.name, type(dep))
            return True
    return False



class DependencyIndex(object):

    """Keeps track of what dependencies belong to what class and provides
    efficient lookups of what task produces what dependency. Use
    ``DependencyIndex[dependency_obj]`` to get the task that makes that
    dependency.

    """

    def __init__(self):
        self._taskidx = defaultdict(dict)
        self._taskidx.update((k, dict()) for k in six.iterkeys(_singleton_idx))

        
    def link(self, dep, task_or_none):
        """Link a dependency to a task. Used later for lookups with
        __getitem__.

        :param dep: The dependency to track
        :type dep: subclass of :class:`anadama2.tracked.Base`

        :param task_or_none: The task that's supposed to create this
          dependency. Use None if the dependency isn't created by a
          task but exists prior to any tasks running.
        :type task_or_none: :class:`anadama2.Task` or None

        """
        self._taskidx[dep.__class__.__name__][dep.name] = task_or_none


    def __contains__(self, dep):
        return dep.name in self._taskidx[dep.__class__.__name__]


    def __getitem__(self, dep):
        if not isinstance(dep, Base):
            raise TypeError(
                "DependencyIndex can only use subclasses of"
                " Base to perform lookups."
                " Received type `{}'".format(type(dep))
            )
        deptype = dep.__class__.__name__
        dep = _singleton_idx[deptype][dep.name]
        return self._taskidx[deptype][dep.name]



class Base(object):

    """The Dependency object is the tool for specifying task father-child
    relationships. Dependency objects can be used in either
    ``targets`` or ``depends`` arguments of
    :meth:`anadama2.workflow.Workflow.add_task`. Often, these targets or
    dependencies are specified by strings or :class:`anadama2.Task`
    objects and instantiated into the appropriate Base
    subclass with :func:`anadama2.tracked.auto`; this behavior depends on the
    arguments to :meth:`anadama2.workflow.Workflow.add_task`.

    A dependency of the same name will be defined multiple times in
    the normal use of AnADAMA. To make it such that many calls of a
    dependency constructor or use of the same argument to
    :func:`anadama2.tracked.auto` result in the same dependency instance being
    returned, a little bit of black magic involving
    :meth:`anadama2.tracked.Base.__new__` is required. The price for
    such magics is that subclasses of Base must define the
    ``init`` method to initialize the dependency, instead of the more
    commonly used ``__init__`` method. The ``init`` method is only
    called once per dependency, while ``__init__`` is called every
    time any Base sublcasses are
    instantiated. Base subclasses must also define a
    ``key()`` staticmethod. The ``key()`` staticmethod is used to
    lookup already existing instances of that dependency; return
    values from the ``key()`` method must be unique to that
    dependency.

    Unrelated to the quasi-singleton magics above, sublcasses of
    Base must define a ``compare()`` method. See
    :meth:`anadama2.tracked.Base.compare` for more documentation.

    """

    must_preexist = True

    def __new__(cls, key, *args, **kwargs):
        global _singleton_idx
        real_key = cls.key(key)
        maybe_exists = _singleton_idx[cls.__name__].get(real_key, None)
        if maybe_exists:
            return maybe_exists
        else:
            _singleton_idx[cls.__name__][real_key] = dep = object.__new__(cls)
            dep.name = real_key
            dep.init(key, *args, **kwargs)
            return dep


    def __getnewargs__(self):
        return (self.name,)
           
    def temp_files(self):
        """ True if tracked generates temp files """
        return False 

    def init(self, key):
        """Initialize the dependency. Only run once for each new dependency
        object with this ``key()``.

        """
        raise NotImplementedError()
        

    def exists(self):
        """Return whether the thing that this object represents
        exists. Examples include whether a file exists, or whether a
        row in a database exists. This method is used when the
        workflow is not in strict mode and the user tries to depend on
        a Tracked object that isn't the target of another task.

        :returns: bool

        """
        return False

    def compare(self):
        """Produce the iterator that is used to determine if this dependency
        has changed. This method is called twice: once before the
        task's actions have been executed for all of a task's
        dependencies and once after all of a task's actions are
        executed for each of the task's targets.

        :returns: iterator

        """
        raise NotImplementedError()


    @staticmethod
    def key(the_key):
        """Returns the unique key for retrieving this dependency from the
        storage backend and for comparing against other dependencies
        of the same type.

        :rtype: str

        """
        raise NotImplementedError()


    def __hash__(self):
        return hash(self.name)



class TrackedString(Base):

    must_preexist = False

    def init(self, s):
        """
        Initialize the dependency.

        :param s: The string to keep track of
        :type s: str or unicode

        """
        self.s = s


    def compare(self):
        yield self.s


    @staticmethod
    def key(s):
        return s

    def __str__(self):
        return self.s



KVDEPSEPARATOR = ":"
class TrackedVariable(Base):
    must_preexist = False

    def __new__(cls, namespace, key, val):
        global _singleton_idx
        real_key = cls.key(namespace, key, val)
        maybe_exists = _singleton_idx[cls.__name__].get(real_key, None)
        if maybe_exists:
            return maybe_exists
        else:
            _singleton_idx[cls.__name__][real_key] = dep = object.__new__(cls)
            dep.name = real_key
            dep.init(namespace, key, val)
            return dep


    def init(self, namespace, k, v):
        self.val = str(v)

    def compare(self):
        yield self.val

    @staticmethod
    def key(ns, k, v):
        return KVDEPSEPARATOR.join(map(str, (ns,k)))

    def __getnewargs__(self):
        ns, k = self.name.split(KVDEPSEPARATOR)
        return (ns, k, self.val)

    def __str__(self):
        return str(self.val)


class TrackedFile(Base):
    """Track a small file. Small being small enough that you don't mind
    that the entire file is read to create a checksum of its contents.

    This dependency class is the safest option for tracking file
    changes. Safety comes at the cost of disk IO; all file contents
    are read to create a checksum.

    """

    def init(self, name):
        """
        Initialize the dependency.
        :param name: The filename to keep track of
        :type name: str
        """
        self.name = self.__class__.key(name)

    def exists(self):
        return os.path.exists(self.name)

    def compare(self):
        stat = os.stat(self.name)
        yield stat.st_size
        yield stat.st_mtime
        yield _adler32(self.name)


    @staticmethod
    def key(name):
        return os.path.abspath(name)

    def __str__(self):
        return self.name



class HugeTrackedFile(TrackedFile):
    """Track a large file. Large being large enough that you don't want to
    read through the entire file to create a checksum of its contents.

    This dependency class is the fastest option for tracking file
    changes for large files. Speed comes at the cost of safety; only
    the size and the modification time are used to determine
    freshness.

    """

    def compare(self):
        stat = os.stat(self.name)
        yield stat.st_size
        yield stat.st_mtime

def try_set_local_path(target, tmpdir):
    """ Try to set the local path for a tracked object """
    try:
        return target.set_local_path(tmpdir)
    except AttributeError:
        return target

def try_get_local_path(target):
    """ Try to get the local path for a tracked object """
    try:
        return target.local
    except AttributeError:
        return target

def download_files_if_needed(depends):
    """ From the list of dependencies, for any that are remote, download if needed """

    for depend in depends:
        try:
            depend.download()
        except AttributeError as err:
            next

def create_temp_folders_if_needed(targets):
    """ Create temp folders for targets if needed """

    for target in targets:
        try:
            target.create_temp_folder()
        except AttributeError as err:
            next

def upload_files_if_needed(targets):
    """ From the list of targets, for any that are remote, upload if needed """

    for target in targets:
        try:
            target.upload()
        except AttributeError as err:
            next

def s3_folder(path):
    """ Check if file/folder is in AWS s3 """
    return True if path and path.startswith("s3://") else False

def s3_bucket(path):
    """ Get the bucket name """
    return path.replace("s3://","").split("/")[0]

def s3_build_path(bucket,key):
    """ Build the full path to the file """
    return "s3://{}/{}".format(bucket,key)

class AWSHugeTrackedFile(HugeTrackedFile):
    """Track a file in the AWS S3 bucket """

    def __init__(self,name):
        import boto3
        self.name = self.__class__.key(name)
        self.aws_bucket = name.replace("s3://","").split("/")[0] 
        self.aws_key = "/".join(name.replace("s3://","").split("/")[1:])
        self.local_base = None
        self.local = None
        self.tmpdir = None

    def temp_files(self):
        return True

    def exists(self):
        import botocore
        import boto3
        resource = boto3.resource("s3")
        found = True
        try:
            last_modified = resource.ObjectSummary(self.aws_bucket,self.aws_key).last_modified
        except botocore.exceptions.ClientError:
            found = False
        return found

    def compare(self):
        import boto3
        resource = boto3.resource("s3")
        aws_object = resource.ObjectSummary(self.aws_bucket,self.aws_key)
        yield aws_object.size
        yield str(aws_object.last_modified)

    @staticmethod
    def key(name):
        return name

    def set_local_path(self, tmpdir):
        self.tmpdir = os.path.split(tmpdir)[-1]
        self.local_base = os.path.join(os.path.abspath(tmpdir),"s3")
        self.local = os.path.join(self.local_base,self.name.replace("s3://",""))
        return self

    def prepend_local_path(self, tmpdir):
        self.local_base = os.path.join(tmpdir, self.local_base[1:] if self.local_base.startswith("/") else self.local_base)
        self.local = os.path.join(tmpdir, self.local[1:] if self.local.startswith("/") else self.local)
        return self.local

    def download(self):
        import boto3
        # create download folder if needed
        directory = os.path.dirname(self.local)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        resource = boto3.resource("s3")
        resource.Bucket(self.aws_bucket).download_file(self.aws_key,self.local)

    def upload(self):
        import boto3
        resource = boto3.resource("s3")
        resource.Bucket(self.aws_bucket).upload_file(self.local,self.aws_key)

    def create_temp_folder(self):
        # create all temp folders needed for local path
        directory = os.path.dirname(self.local)
        if not os.path.isdir(directory):
            os.makedirs(directory)

class Container(object):
    """Track a collection of small strings. This is useful for rerunning
    tasks based on whether a flag has changed when running a
    command. Using :class:`anadama2.tracked.TrackedString` for small
    strings can run into collisions between workflows that use the
    same backend. Consider using ``logging =
    TrackedString("debug")`` in script_a.py and ``logging =
    TrackedString("warning")`` in script_b.py. If you change
    script_a.py to ``logging = TrackedString("warning")`` after
    running script_b.py, script_a.py won't rerun tasks that depend on
    the TrackedString assigned to ``logging``.

    This class solves TrackedString collisions by prepending a
    user-provided (or auto-generated) namespace to each key-value pair
    in the collection. The auto-generated namespace is unique to the
    script or module creating the Container, so expect to have
    a bunch of tasks rerun if you rename a module or script between
    runs.

    .. code:: python

      >>> import anadama2.tracked
      >>> conf = anadama2.tracked.Container(alpha="5", beta=2)
      >>> conf.alpha
      <anadama2.tracked.TrackedVariable object at 0x7f66445fa490> 
      >>> str(conf.alpha)
      '5'
      >>> str(conf['beta'])
      '2'
      >>> conf.beta = 7
      >>> conf.beta
      <anadama2.tracked.TrackedVariable object at 0x7f66445fa4d0>
      >>> str(conf.beta)
      '7'
    
    """

    def __init__(self, namespace=None, **kwds):
        self.__dict__['_ns'] = self.__class__.key(namespace)
        logger.debug("Creating %s with namespace %s",
                     self.__class__.__name__, self._ns)
        self.__dict__['_d'] = dict()
        for k, v in kwds.items():
            self._d[k] = TrackedVariable(self._ns, k, v)

    def temp_files(self):
        return False

    @staticmethod
    def key(namespace=None):
        if not namespace:
            return os.path.abspath(os.path.dirname(sys.argv[0]))
        return namespace

    def items(self):
        return list(self._d.values())

    def compare(self):
        for k in self._d:
            for item in self[k].compare():
                yield item


    def __getattr__(self, key):
        return self._d[key]

    __getitem__ = __getattr__

    def __setattr__(self, key, val):
        if key in self._d:
            global _singleton_idx
            dep = self._d[key]
            dep.val = val
            dep.name = TrackedVariable.key(self._ns, key, val)
            _singleton_idx[dep.__class__.__name__][dep.name] = dep
        else:
            self._d[key] = TrackedVariable(self._ns, key, val)

    __setitem__ = __setattr__

    def __hash__(self):
        return hash(tuple(self._d.items()))



class TrackedDirectory(TrackedFile):

    """Track a directory. A directory is considered changed if it's
    removed, the modify time has changed, the list of files within the
    directory has changed, or if any of the files within the directory
    have changed size or modify times.

    """

    def exists(self):
        return os.path.isdir(self.name)

    def compare(self):
        stat = os.stat(self.name)
        yield stat.st_size
        yield stat.st_mtime
        contained = list(sorted(os.listdir(self.name)))
        yield hash(tuple(contained))
        for item in contained:
            stat = os.stat(os.path.join(self.name, item))
            yield stat.st_size
            yield stat.st_mtime



class TrackedFilePattern(TrackedFile):
    """Track several files according to a bash-style globbing
    pattern. Uses :func:`glob.glob` under the hood.  A Glob is
    considered changed if the names of the matched files changes, or
    any of the matched files change in size or modify time.

    """

    def exists(self):
        return bool(glob(self.name))

    def compare(self):
        fs = glob(self.name)
        fs = list(sorted(fs))
        yield hash(tuple(fs))
        for f in fs:
            stat = os.stat(f)
            yield stat.st_size
            yield stat.st_mtime



class TrackedExecutable(Base):
    """Track a script or binary executable."""

    def init(self, name, version_command="{} --version"):
        """Initialize the dependency.

        :param name: Name of a script on the shell $PATH or name of
          the file to track
        :type name: str
        
        :keyword version: Command to get the executables version
        :type version: str
        """

        self.name = self.__class__.key(name)
        self.version_command=version_command.format(name)
            
    def version(self):
        try:
            version = subprocess.check_output(self.version_command, shell=True, stderr=subprocess.STDOUT)
        except (subprocess.CalledProcessError, EnvironmentError):
            version = None
        
        return version

    def exists(self):
        return os.path.exists(self.name)
        

    def compare(self):
        version = self.version()
        if version:
            yield version
        stat = os.stat(self.name)
        yield stat.st_size
        yield stat.st_mtime
        yield _adler32(self.name)


    @staticmethod
    def key(name):
        # if the current path exists, it is a file (not a directory), and it is executable
        # then use the existing path
        if os.path.exists(name) and os.path.isfile(name) and os.access(name, os.X_OK):
            p = name
        else:
            p = find_on_path(name)
            if not p:
                raise ValueError(
                    "Unable to find binary or script `{}'".format(name))
        return p

    def __str__(self):
        return self.name



class TrackedFunction(Base):

    """Useful for things like database lookups or API calls. The function
    must return a hashable type. For a tiered comparison method like
    that seen in :class:`anadama2.tracked.TrackedFile`, it's best to create
    your own subclass of Base and override the ``compare()``
    method.

    """

    must_preexist = False

    def init(self, key, fn):
        self.fn = fn
        self.name = key

    def compare(self):
        yield self.fn()


    @staticmethod
    def key(key):
        return key



_cached_dep_classes = (
    Base,               TrackedString,
    TrackedFile,        HugeTrackedFile,
    TrackedExecutable,  TrackedFunction,
    TrackedFilePattern, TrackedDirectory,
    TrackedVariable
)
for cls in _cached_dep_classes:
    _singleton_idx[cls.__name__] = dict()
del cls



import os
import sys
import logging
import multiprocessing
import collections

import six
import string
import time

from .util import mkdirp
from .tracked import TrackedExecutable, s3_folder

LOG_FILE_NAME = "anadama.log"

def default(output_dir=None, log_level=None):
    log = LOG_FILE_NAME
    if not output_dir or s3_folder(output_dir):
        output_dir=os.getcwd()
    log = os.path.abspath(os.path.join(output_dir, log))
    return ReporterGroup([
        LoggerReporter(log_level, log),
        VerboseConsoleReporter()
    ])

class BaseReporter(object):

    """The base reporter defines functionality similar amongst all
    reporters.

    The workflow that executes the hooks defined can be accessed at
    ``self.run_context``.
    """

    def started(self, run_context):
        """Executed when a run is started, usually when
        :meth:`anadama2.workflow.Workflow.go` is executed.

        """
        self.run_context = run_context
        raise NotImplementedError()

    def task_skipped(self, task_no):
        """Executed when anadama determines that a task needn't be run.

        :param task_no: The task number of the task that is
          skipped. To get the actual :class:`anadama2.Task` object
          that's being skipped, do ``self.run_context.tasks[task_no]``.

        :type task_no: int

        """

        raise NotImplementedError()

    def task_started(self, task_no):
        """Executed when anadama is just about to execute a task. These tasks
        are in the queue and waiting for available resources.

        :param task_no: The task number of the task that is
          being started. To get the actual :class:`anadama2.Task` object
          that's being executed, do ``self.run_context.tasks[task_no]``.

        :type task_no: int

        """

        raise NotImplementedError()

    def task_running(self, task_no):
        """Executed when anadama is ready to run a task.

        :param task_no: The task number of the task that is
          being started. To get the actual :class:`anadama2.Task` object
          that's being executed, do ``self.run_context.tasks[task_no]``.

        :type task_no: int

        """

        raise NotImplementedError()
    
    def task_command(self, task_no):
        """Executed when anadama is ready to run a task. Logs the command
            and version of any tracked executables.

        :param task: The task number of the task that is
          being started. To get the actual :class:`anadama2.Task` object
          that's being executed, do ``self.run_context.tasks[task_no]``.

        :type task_no: int

        """

        pass

    def task_failed(self, task_result):
        """Executed when a task fails.

        :param task_no: The task number of the task that failed . To
          get the actual :class:`anadama2.Task` object that failed, do
          ``self.run_context.tasks[task_no]``. To get the task result
          of the task that failed,
          do``self.run_context.task_results[task_no]``

        :type task_no: int

        """
        raise NotImplementedError()        

    def task_completed(self, task_result):
        """Executed when a task completes with no errors.

        :param task_no: The task number of the task that succeeded . To
          get the actual :class:`anadama2.Task` object that succeeded, do
          ``self.run_context.tasks[task_no]``. To get the task result
          of the task that succeeded,
          do``self.run_context.task_results[task_no]``

        :type task_no: int

        """
        raise NotImplementedError()

    def finished(self):
        """Executed when a run finishes. This method is called whether there
        are task failures or not.

        """
        raise NotImplementedError()

    def task_grid_status(self, task_no, grid_id, status_message):
        """Executed when anadama has grid information for a task. These messages 
        are reported when the status for a grid task has changed.

        :param task_no: The task number of the task that is
          being started. To get the actual :class:`anadama2.Task` object
          that's being executed, do ``self.run_context.tasks[task_no]``.
        :param grid_id: The id of the grid job.
        :param status_message: The grid status message.

        :type task_no: int

        """

        raise NotImplementedError()
    
    def task_grid_status_polling(self, task_no, grid_id, status_message):
        """Executed when anadama has grid information for a task at set polling 
        intervals. Status may repeat at each interval.

        :param task_no: The task number of the task that is
          being started. To get the actual :class:`anadama2.Task` object
          that's being executed, do ``self.run_context.tasks[task_no]``.
        :param grid_id: The id of the grid job.
        :param status_message: The grid status message.

        :type task_no: int

        """

        raise NotImplementedError()


class ReporterGroup(BaseReporter):
    """Sometimes you want to use multiple reporters. For that, there is
    ReporterGroup. Here's an example usage:

    .. code:: python

      from anadama2.reporters import ReporterGroup
      my_grouped_reporter = ReporterGroup([custom_reporter_a, 
                                           custom_reporter_b, 
                                           custom_reporter_c])
      ...
      ctx.go(reporter=my_grouped_reporter)

    """

    def __init__(self, other_reporters):
        self.reps = other_reporters


    def started(self, ctx):
        for r in self.reps:
            r.started(ctx)


    def task_skipped(self, task_no):
        for r in self.reps:
            r.task_skipped(task_no)


    def task_started(self, task_no):
        for r in self.reps:
            r.task_started(task_no)

    def task_running(self, task_no):
        for r in self.reps:
            r.task_running(task_no)
            
    def task_command(self, task_no):
        for r in self.reps:
            r.task_command(task_no)

    def task_failed(self, task_result):
        for r in self.reps:
            r.task_failed(task_result)


    def task_completed(self, task_result):
        for r in self.reps:
            r.task_completed(task_result)


    def finished(self):
        for r in self.reps:
            r.finished()

    def task_grid_status(self, task_no, grid_id, status_message):
        for r in self.reps:
            r.task_grid_status(task_no, grid_id, status_message)

    def task_grid_status_polling(self, task_no, grid_id, status_message):
        for r in self.reps:
            r.task_grid_status_polling(task_no, grid_id, status_message)

class ConsoleReporter(BaseReporter):
    """Prints out run progress to stderr.
    An example readout is as follows:

    ::

      (s)[  1/  6 -  16.67%] Track pre-existing dependencies

    The readout is composed of five pieces of information:

      1. The task status. That's the part in the parentheses. 
      
        * ``( )`` means that the task is currently being executed

        * ``(+)`` means that the task finished successfully

        * ``(s)`` means that the task was skipped

        * ``(!)`` means that the task failed.

      2. The current task number. That's the first number in the
         square brackets.

      3. The total number of tasks to be run or skipped. That's the
         number after the forward slash.

      4. The percent complete of the current run. That's the number
         with a percent-sign next to it

      5. The task name. That's the text that comes after the ending
         square bracket. Remember that you can set the task name with
         the ``name`` option to
         :meth:`anadama2.workflow.Workflow.add_task`.

    """


    msg_str = six.u("({:.1})[{:3}/{:3} - {:6.2f}%] {:.57}")

    class stats:
        skip  = six.u("s")
        fail  = six.u("!")
        done  = six.u("+")
        start = six.u(" ")

    def __init__(self, *args, **kwargs):
        self.failed_results = list()
        self.n_open = 0
        self.multithread_mode = False


    def _msg(self, status, msg, c_r=False, visible=True):
        if self.n_open > 1 and not self.multithread_mode:
            self.multithread_mode = True
            if visible is True:
                sys.stderr.write(six.u('\n'))
        s = self.msg_str.format(status, self.n_complete, self.n_tasks,
                                (float(self.n_complete)/self.n_tasks)*100,
                                six.u(msg))
        if self.multithread_mode is True:
            s += six.u("\n")
        elif c_r:
            self.n_open -= 1
            s = six.u("\r") + s + six.u("\n")
        else:
            self.n_open += 1
        if visible is True:
            sys.stderr.write(s)


    def started(self, ctx):
        self.run_context = ctx
        self.reset()

    def task_started(self, task_no):
        self._msg(self.stats.start, self.run_context.tasks[task_no].name,
                  visible=self.run_context.tasks[task_no].visible)
    
    def task_skipped(self, task_no):
        self.n_complete += 1
        self._msg(self.stats.skip, self.run_context.tasks[task_no].name, True,
                  visible=self.run_context.tasks[task_no].visible)

    def task_failed(self, task_result):
        self.n_complete += 1
        if task_result.task_no is None:
            return
        name = self.run_context.tasks[task_result.task_no].name
        self.failed_results.append((name, task_result))
        self._msg(self.stats.fail, name, True,
                  visible=self.run_context.tasks[task_result.task_no].visible)

    def task_completed(self, task_result):
        self.n_complete += 1
        name = self.run_context.tasks[task_result.task_no].name
        self._msg(self.stats.done, name, True,
                  visible=self.run_context.tasks[task_result.task_no].visible)

    def finished(self):
        sys.stderr.write(six.u("Run Finished\n"))
        for name, result in self.failed_results:
            sys.stderr.write(six.u("Task {} failed\n".format(result.task_no)))
            sys.stderr.write(six.u("  Name: "+name+"\n"))
            sys.stderr.write(six.u("  Original error: \n"))
            for line in result.error.split("\n"):
                sys.stderr.write(six.u("  "+line+"\n"))
        self.reset()

    def reset(self):
        self.n_tasks = len(self.run_context.tasks)
        self.n_complete = 0
        self.failed = False

class VerboseConsoleReporter(BaseReporter):
    """Prints out verbose run progress to stdout.
    An example readout is as follows:

    ::

    DATE/TIME [ 0/18 -   0.00%] **Started  ** Task  2: kneaddata
    DATE/TIME [ 0/18 -   0.00%] **Started  ** Task  0: kneaddata
    DATE/TIME [ 1/18 -   5.56%] **Completed** Task  0: kneaddata
    DATE/TIME [ 1/18 -   5.56%] **Started  ** Task  4: metaphlan2.py
    DATE/TIME [ 2/18 -  11.11%] **Completed** Task  2: kneaddata

    The readout is composed of five pieces of information:
    
      1. The date/time for the status message.
      2. The status of all tasks. For example, [1/4 - 25%] indicates that 
          one task of the four total tasks have finished running. The
          workflow is 25% complete.
      3. The step the task has completed. Examples are "Started" and "Completed".
      4. The task number. 
      5. The task description. This is the task name if set. If the task name is
          the default then it is the first task action. This is the first command
          and it is limited to the executable name. If it is a function, it will be the 
          name of the function.
    """

    class stats:
        skip  = six.u("Skipped")
        fail  = six.u("Failed")
        done  = six.u("Completed")
        start = six.u("Started")
        ready = six.u("Ready")
        grid_run = six.u("GridJob")
        max_message_length = max(len(x) for x in [skip,fail,done,start,ready,grid_run])

    def __init__(self, *args, **kwargs):
        self.failed_results = list()
        # Use a multiprocessing value so the total tasks completed is shared
        # for the local multiprocessing tasks (is process and thread-safe)
        self.n_complete = multiprocessing.Value('i',0)
        
        # Set the max length for the total output line
        self.max_length=120

    def _msg(self, status, task_name, description, id, visible=True, grid_update=None):
        # create a date/time string
        s = time.strftime("(%b %d %H:%M:%S) ", time.localtime())
        
        # create a message string for the current status
        s += self.msg_str.format(self.n_complete.value, self.n_tasks,
                                (float(self.n_complete.value)/self.n_tasks)*100, 
                                status, id, description)
        # if command string is reduced, add ellipses
        if len(description) > self.max_command_length:
            s += " ..."
            
        if grid_update:
            s += self.grid_update_msg.format(grid_update[0], grid_update[1])
            
        s += six.u("\n")
        
        # only write if the task is visible
        if visible is True:
            sys.stdout.write(s)
            
    def _increment_complete(self, task_no):
        # update the number of completed visible tasks
        if self.run_context.tasks[task_no].visible:
            self.n_complete.value+=1

    def started(self, ctx):
        self.run_context = ctx
        self.reset()

    def task_started(self, task_no):
        # This indicates the task is ready and waiting in the queue
        self._msg(self.stats.ready, self.run_context.tasks[task_no].name,
                  self.run_context.tasks[task_no].description, task_no,
                  visible=self.run_context.tasks[task_no].visible)
        
    def task_running(self, task_no):
        self._msg(self.stats.start, self.run_context.tasks[task_no].name,
                  self.run_context.tasks[task_no].description, task_no,
                  visible=self.run_context.tasks[task_no].visible)
    
    def task_skipped(self, task_no):
        self._increment_complete(task_no)
        self._msg(self.stats.skip, self.run_context.tasks[task_no].name, 
                  self.run_context.tasks[task_no].description, task_no,
                  visible=self.run_context.tasks[task_no].visible)

    def task_failed(self, task_result):
        self._increment_complete(task_result.task_no)
        if task_result.task_no is None:
            return
        name = self.run_context.tasks[task_result.task_no].name
        self.failed_results.append((name, task_result))
        self._msg(self.stats.fail, name, self.run_context.tasks[task_result.task_no].description,
                  task_result.task_no, visible=self.run_context.tasks[task_result.task_no].visible)

    def task_completed(self, task_result):
        self._increment_complete(task_result.task_no)
        name = self.run_context.tasks[task_result.task_no].name
        self._msg(self.stats.done, name, self.run_context.tasks[task_result.task_no].description,
                  task_result.task_no, visible=self.run_context.tasks[task_result.task_no].visible)

    def task_grid_status(self, task_no, grid_id, status_message):
        self._msg(self.stats.grid_run, self.run_context.tasks[task_no].name,
                  self.run_context.tasks[task_no].description, task_no, visible=True,
                  grid_update=[grid_id, status_message])
        
    def task_grid_status_polling(self, task_no, grid_id, status_message):
        self.task_grid_status(task_no, grid_id, status_message)

    def finished(self):
        sys.stdout.write(six.u("Run Finished\n"))
        for name, result in self.failed_results:
            sys.stdout.write(six.u("Task {} failed\n".format(result.task_no)))
            sys.stdout.write(six.u("  Name: "+name+"\n"))
            sys.stdout.write(six.u("  Original error: \n"))
            for line in result.error.split("\n"):
                print_line="".join(filter(lambda x: x in string.printable, line))
                sys.stdout.write(six.u("  "+print_line+"\n"))
        self.reset()

    def reset(self):
        # count the number of visible tasks
        self.n_tasks = 0
        for task in self.run_context.tasks:
            if task.visible:
                self.n_tasks+=1
                
        # limit the full string length
        max_task_length=len(str(self.n_tasks))
        self.max_command_length = self.max_length -(20+max_task_length*3+20+self.stats.max_message_length)
        self.msg_str = six.u("[{:"+str(max_task_length)+"}/{:"+str(max_task_length)+
                             "} - {:6.2f}%] **{:"+str(self.stats.max_message_length)+
                             "}** Task {:"+str(max_task_length)+
                             "}: {:."+str(self.max_command_length)+"}")
        self.grid_update_msg = six.u(" <Grid JobId {:9}: {:.30}>")
        self.n_complete.value = 0
        self.failed = False
        
SHELL_COMMAND = "Executing with shell: "
VERSION_COMMAND = "Tracked executable version: "

class LoggerReporter(BaseReporter):
    """A reporter that uses :mod:`logging`.

    :param loglevel_str: The logging level. Valid levels: subdebug,
      debug, info, subwarning, warning, error
    :type loglevel_str: str

    :param logfile: The file to log to. Defaults to stdout.
    :type logfile: str or file-like

    :param fmt_str: The log format. See :mod:`logging` for more
      information
    :type fmt_str: str

    """

    FORMAT = "%(asctime)s\t%(name)s\t%(funcName)s\t%(levelname)s: %(message)s"
    
    def __init__(self, loglevel_str=None, logfile=None,
                 fmt_str=None, *args, **kwargs):
        # create the log file folder if needed
        if logfile:
            mkdirp(os.path.dirname(logfile))        
        self.logger = logging.getLogger(self.__class__.__name__)
        self.loglevel_str = loglevel_str.upper() or logging.WARNING
        loglevel = getattr(logging, self.loglevel_str)
        logkwds = {"format": fmt_str or self.FORMAT,
                  "level":  loglevel }
        if logfile and hasattr(logfile, "write"):
            logkwds['stream'] = logfile
        elif logfile and isinstance(logfile, six.string_types):
            logkwds['filename'] = logfile
        logging.basicConfig(**logkwds)
        self.any_failed = False
        self.start_log_message="task %i, %s : %s "
        
    @classmethod
    def read_log(cls,file,type,remove_paths=True):
        """ Read the data from the log file """
        
        # read all of the lines from the log
        data=collections.OrderedDict()
        with open(file) as file_handle:
            lines=file_handle.readlines()
        
        # look for commands, benchmarking, or executable version information
        log_info={}
        if type == "commands":
            keyword=SHELL_COMMAND
            for line in lines:
                if keyword in line:
                    new_command=line.split(keyword)[-1].strip()
                    if remove_paths:
                        new_command=" ".join([os.path.split(i.rstrip(os.path.sep))[-1] for i in new_command.split(" ")])
                    data[new_command]=1
            log_info=list(data.keys())
        elif type == "benchmarking":
            benchmarking_info={}
            # read in the commands and also the benchmarking information
            for i in range(len(lines)):
                if "run_task_command" in lines[i]:
                    # get the id and the executable plus input item
                    id=lines[i].split()[-1].strip().replace(":","")
                    command_tokens=lines[i+1].strip().split()
                    executable=command_tokens[0]
                    try:
                        input=os.path.split(command_tokens[command_tokens.index("--input")+1])[-1]
                    except (ValueError, IndexError):
                        input="NA"
                    log_info[id]="\t".join([executable, input])
                elif "Benchmark" in lines[i]:
                    # get the job id and benchmarking info
                    id=lines[i].split()[-1].strip().replace(":","")
                    info="\t".join([lines[j].strip().split(": ")[-1] for j in [i+1,i+2,i+3]])
                    benchmarking_info[id]=info
                    
            # set the log info to include id, executable, input and benchmarking
            for id in benchmarking_info.keys():
                if id in log_info:
                    log_info[id]=log_info[id]+"\t"+benchmarking_info[id]
        elif type == "variables":
            # read in the variables set for the workflow
            # use the last variable values for workflow with more than one run
            for line in lines:
                if "started\tINFO:" in line and " = " in line:
                    data = line.rstrip().split(": ")[-1]
                    try:
                        variable, value = data.split(" = ")
                        log_info[variable]=value
                    except ValueError:
                        pass
        else:
            keyword = VERSION_COMMAND
            # remove redundant version from strings since these are already 
            # identified as version information
            format_output=lambda x: x.replace("Version:","").replace(", version","")
            for line in lines:
                if keyword in line:
                    data[format_output(line.split(keyword)[-1].strip())]=1
            log_info=list(data.keys())
                    
        if not log_info:
            log_info=["No {} found in log".format(type)]
                
        return log_info

    def _daginfo(self, task_no):
        children = self.run_context.dag.successors(task_no)
        parents = self.run_context.dag.predecessors(task_no)
        msg = " {} parents: {}.  {} children: {}."
        return msg.format(len(parents), parents, len(children), children)
    
    def log_event(self, msg, task_no, debug_msg=None):
        visible=self.run_context.tasks[task_no].visible
        description=self.run_context.tasks[task_no].description
        
        # If in debug level and debug message is provided, append to end of message
        if self.loglevel_str == "DEBUG" and debug_msg is not None:
            msg=msg+" : "+debug_msg
        
        # Write tasks that are not visible to log on debug level
        if visible:
            self.logger.info(self.start_log_message, task_no, description, msg)
        else:
            self.logger.debug(self.start_log_message, task_no, description, msg)

    def started(self, ctx):
        self.run_context = ctx
        self.logger.info("Beginning AnADAMA run with %i tasks.",
                         len(self.run_context.tasks))
        
        # if there are workflow variables set, add them to the log
        try:
            options=vars(ctx.vars.get_option_values())
        except (AttributeError, TypeError):
            options={}
            
        if options:
            self.logger.info("Workflow configuration options")
        
        for name, value in options.items():
            self.logger.info("{} = {}".format(name, value))
            

    def task_skipped(self, task_no):
        self.log_event("skipped", task_no, self._daginfo(task_no))

    def task_started(self, task_no):
        self.log_event("ready and waiting for resources", task_no, self._daginfo(task_no))
        
    def task_running(self, task_no):
        self.log_event("starting to run",task_no,self._daginfo(task_no))
        
    def task_command(self, task_no):
        # if a tracked executable is found, then log the version
        for exe_depends in filter(lambda x: isinstance(x, TrackedExecutable), self.run_context.tasks[task_no].depends):
            version = exe_depends.version()
            if version:
                self.logger.info(VERSION_COMMAND+" "+version)
        
        # if a command, then log the shell command(s)
        if not list(filter(lambda x: six.callable(x), self.run_context.tasks[task_no].actions)):
            self.logger.info(SHELL_COMMAND+" "+" ".join(self.run_context.tasks[task_no].actions))

    def task_failed(self, task_result):
        self.logger.error(self.start_log_message,task_result.task_no,
                          self.run_context.tasks[task_result.task_no].description,
                          " Failed! Error message : {}".format(task_result.error))
        self.any_failed = True

    def task_grid_status(self, task_no, grid_id, status_message):
        self.log_event(" grid job id {} has status {}".format(grid_id, status_message),
            task_no, self.run_context.tasks[task_no].description)
        
    def task_grid_status_polling(self, task_no, grid_id, status_message):
        self.logger.debug(self.start_log_message,
            task_no, self.run_context.tasks[task_no].description, 
            " grid job id {} has status {}".format(grid_id, status_message))

    def task_completed(self, task_result):
        self.log_event("completed successfully",task_result.task_no)

    def finished(self):
        if self.any_failed:
            self.logger.error("AnADAMA run finished with errors.")
        else:
            self.logger.info("AnADAMA run finished.")

        



class WebhookReporter(BaseReporter):
    """TODO"""
    pass

